from __future__ import annotations

import json
import unittest
from contextlib import nullcontext
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, Mock, call, patch
from urllib.parse import parse_qsl, urlparse

import pytest

from streamlink.exceptions import NoStreamsError, PluginError
from streamlink.options import Options
from streamlink.plugin.api import useragents
from streamlink.plugins.twitch import Twitch, TwitchAPI, TwitchHLSStream, TwitchHLSStreamReader, TwitchHLSStreamWriter
from tests.mixins.stream_hls import EventedHLSStreamWriter, Playlist, Segment as _Segment, Tag, TestMixinStreamHLS
from tests.plugins import PluginCanHandleUrl
from tests.resources import text


if TYPE_CHECKING:
    import requests_mock as rm

    from streamlink import Streamlink


class TestPluginCanHandleUrlTwitch(PluginCanHandleUrl):
    __plugin__ = Twitch

    should_match_groups = [
        (
            ("live", "https://www.twitch.tv/CHANNELNAME"),
            {
                "channel": "CHANNELNAME",
            },
        ),
        (
            ("live", "https://www.twitch.tv/CHANNELNAME?"),
            {
                "channel": "CHANNELNAME",
            },
        ),
        (
            ("live", "https://www.twitch.tv/CHANNELNAME/"),
            {
                "channel": "CHANNELNAME",
            },
        ),
        (
            ("live", "https://www.twitch.tv/CHANNELNAME/?"),
            {
                "channel": "CHANNELNAME",
            },
        ),
        (
            ("vod", "https://www.twitch.tv/videos/1963401646"),
            {
                "video_id": "1963401646",
            },
        ),
        (
            ("vod", "https://www.twitch.tv/dota2ti/v/1963401646"),
            {
                "video_id": "1963401646",
            },
        ),
        (
            ("vod", "https://www.twitch.tv/dota2ti/video/1963401646"),
            {
                "video_id": "1963401646",
            },
        ),
        (
            ("vod", "https://www.twitch.tv/videos/1963401646?t=1h23m45s"),
            {
                "video_id": "1963401646",
            },
        ),
        (
            ("clip", "https://clips.twitch.tv/GoodEndearingPassionfruitPMSTwin-QfRLYDPKlscgqt-4"),
            {
                "clip_id": "GoodEndearingPassionfruitPMSTwin-QfRLYDPKlscgqt-4",
            },
        ),
        (
            ("clip", "https://www.twitch.tv/clip/GoodEndearingPassionfruitPMSTwin-QfRLYDPKlscgqt-4"),
            {
                "clip_id": "GoodEndearingPassionfruitPMSTwin-QfRLYDPKlscgqt-4",
            },
        ),
        (
            ("clip", "https://www.twitch.tv/lirik/clip/GoodEndearingPassionfruitPMSTwin-QfRLYDPKlscgqt-4"),
            {
                "clip_id": "GoodEndearingPassionfruitPMSTwin-QfRLYDPKlscgqt-4",
            },
        ),
        (
            ("clip", "https://twitch.tv/papaplatte/clip/SmellyDeadMomBloodTrail-WWr5gMxd0pe0BAge"),
            {
                "clip_id": "SmellyDeadMomBloodTrail-WWr5gMxd0pe0BAge",
            },
        ),
        (
            ("player", "https://player.twitch.tv/?parent=twitch.tv&channel=CHANNELNAME"),
            {},
        ),
        (
            ("player", "https://player.twitch.tv/?parent=twitch.tv&video=1963401646"),
            {},
        ),
        (
            ("player", "https://player.twitch.tv/?parent=twitch.tv&video=1963401646&t=1h23m45s"),
            {},
        ),
    ]

    should_not_match = [
        "https://www.twitch.tv",
        "https://www.twitch.tv/",
        "https://www.twitch.tv/videos/",
        "https://www.twitch.tv/dota2ti/v",
        "https://www.twitch.tv/dota2ti/video/",
        "https://clips.twitch.tv/",
        "https://www.twitch.tv/clip/",
        "https://www.twitch.tv/lirik/clip/",
        "https://player.twitch.tv/",
        "https://player.twitch.tv/?",
    ]


DATETIME_BASE = datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc)
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


class TagDateRangeAd(Tag):
    def __init__(self, start=DATETIME_BASE, duration=1, attrid="stitched-ad-1234", classname="twitch-stitched-ad", custom=None):
        attrs = {
            "ID": self.val_quoted_string(attrid),
            "CLASS": self.val_quoted_string(classname),
            "START-DATE": self.val_quoted_string(start.strftime(DATETIME_FORMAT)),
            "DURATION": duration,
        }
        if custom is not None:
            attrs.update(**{key: self.val_quoted_string(value) for (key, value) in custom.items()})
        super().__init__("EXT-X-DATERANGE", attrs)


class Segment(_Segment):
    def __init__(self, num, title="live", *args, **kwargs):
        super().__init__(num, title, *args, **kwargs)
        self.date = DATETIME_BASE + timedelta(seconds=num)

    def build(self, namespace):
        return f"#EXT-X-PROGRAM-DATE-TIME:{self.date.strftime(DATETIME_FORMAT)}\n{super().build(namespace)}"


class SegmentPrefetch(Segment):
    def build(self, namespace):
        return f"#EXT-X-TWITCH-PREFETCH:{self.url(namespace)}"


class _TwitchHLSStreamWriter(EventedHLSStreamWriter, TwitchHLSStreamWriter):
    pass


class _TwitchHLSStreamReader(TwitchHLSStreamReader):
    __writer__ = _TwitchHLSStreamWriter


class _TwitchHLSStream(TwitchHLSStream):
    __reader__ = _TwitchHLSStreamReader


def test_stream_weight(requests_mock: rm.Mocker, session: Streamlink):
    plugin = Twitch(session, "http://twitch.tv/foo")

    with text("hls/test_master_twitch_vod.m3u8") as fh:
        playlist = fh.read()

    requests_mock.request(method="GET", url="http://mocked/master.m3u8", text=playlist)
    streams = TwitchHLSStream.parse_variant_playlist(session, "http://mocked/master.m3u8")

    with patch.object(plugin, "_get_streams", return_value=streams):
        data = plugin.streams()

    assert list(data.keys()) == ["audio", "160p30", "360p30", "480p30", "720p30", "720p60", "source", "worst", "best"]
    assert data["best"] is data["source"]
    assert data["worst"] is data["160p30"]


def test_multivariant(session: Streamlink, requests_mock: rm.Mocker):
    with text("hls/test_multivariant_twitch_usher_v2.m3u8") as fd:
        content = fd.read()
        requests_mock.get("http://mocked/multivariant.m3u8", text=content)
        streams = TwitchHLSStream.parse_variant_playlist(session, "http://mocked/multivariant.m3u8")

    assert sorted(streams.keys()) == ["1080p60", "160p", "360p", "480p", "720p60", "audio_only"]


@patch("streamlink.stream.hls.HLSStreamWorker.wait", MagicMock(return_value=True))
class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
    __stream__ = _TwitchHLSStream
    stream: TwitchHLSStream

    def get_session(self, *args, **kwargs):
        session = super().get_session(*args, **kwargs)
        session.set_option("hls-live-edge", 4)

        return session

    def test_hls_daterange_unknown(self):
        daterange = TagDateRangeAd(
            start=DATETIME_BASE,
            duration=1,
            attrid="foo",
            classname="bar",
            custom=None,
        )

        segments = self.subject(
            [
                Playlist(0, [daterange, Segment(0), Segment(1)], end=True),
            ],
            streamoptions={"low_latency": False},
        )

        self.await_write(2)
        data = self.await_read(read_all=True)
        assert data == self.content(segments), "Doesn't filter out segments"
        assert all(self.called(s) for s in segments.values()), "Downloads all segments"

    def test_hls_daterange_by_class(self):
        daterange = TagDateRangeAd(
            start=DATETIME_BASE,
            duration=1,
            attrid="foo",
            classname="twitch-stitched-ad",
            custom=None,
        )

        segments = self.subject(
            [
                Playlist(0, [daterange, Segment(0), Segment(1)], end=True),
            ],
            streamoptions={"low_latency": False},
        )

        self.await_write(2)
        data = self.await_read(read_all=True)
        assert data == segments[1].content, "Filters out ad segments"
        assert all(self.called(s) for s in segments.values()), "Downloads all segments"

    def test_hls_daterange_by_id(self):
        daterange = TagDateRangeAd(
            start=DATETIME_BASE,
            duration=1,
            attrid="stitched-ad-1234",
            classname="/",
            custom=None,
        )

        segments = self.subject(
            [
                Playlist(0, [daterange, Segment(0), Segment(1)], end=True),
            ],
            streamoptions={"low_latency": False},
        )

        self.await_write(2)
        data = self.await_read(read_all=True)
        assert data == segments[1].content, "Filters out ad segments"
        assert all(self.called(s) for s in segments.values()), "Downloads all segments"

    @patch("streamlink.plugins.twitch.log")
    def test_hls_has_preroll(self, mock_log):
        daterange = TagDateRangeAd(
            duration=4,
            custom={"X-TV-TWITCH-AD-ROLL-TYPE": "PREROLL"},
        )
        segments = self.subject(
            [
                Playlist(0, [daterange, Segment(0), Segment(1)]),
                Playlist(2, [daterange, Segment(2), Segment(3)]),
                Playlist(4, [Segment(4), Segment(5)], end=True),
            ],
            streamoptions={"low_latency": False},
        )

        self.await_write(6)
        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: s.num >= 4), "Filters out preroll ad segments"
        assert all(self.called(s) for s in segments.values()), "Downloads all segments"
        assert self.thread.reader.worker.duration == pytest.approx(2.0), "Ad segments don't affect the output duration"
        assert mock_log.info.mock_calls == [
            call("Will skip ad segments"),
            call("Waiting for pre-roll ads to finish, be patient"),
            call("Detected advertisement break of 4 seconds"),
        ]

    @patch("streamlink.plugins.twitch.log")
    def test_hls_has_midroll(self, mock_log):
        daterange = TagDateRangeAd(
            start=DATETIME_BASE + timedelta(seconds=2),
            duration=2,
            custom={"X-TV-TWITCH-AD-ROLL-TYPE": "MIDROLL", "X-TV-TWITCH-AD-COMMERCIAL-ID": "123"},
        )
        segments = self.subject(
            [
                Playlist(0, [Segment(0), Segment(1)]),
                Playlist(2, [daterange, Segment(2), Segment(3)]),
                Playlist(4, [Segment(4), Segment(5)], end=True),
            ],
            streamoptions={"low_latency": False},
        )

        self.await_write(6)
        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: s.num != 2 and s.num != 3), "Filters out mid-stream ad segments"
        assert all(self.called(s) for s in segments.values()), "Downloads all segments"
        assert self.thread.reader.worker.duration == pytest.approx(4.0), "Ad segments don't affect the output duration"
        assert mock_log.info.mock_calls == [
            call("Will skip ad segments"),
            call("Detected advertisement break of 2 seconds"),
        ]

    @patch("streamlink.plugins.twitch.log")
    def test_hls_has_preroll_and_midroll(self, mock_log):
        ads1a = TagDateRangeAd(
            start=DATETIME_BASE,
            duration=2,
            custom={"X-TV-TWITCH-AD-ROLL-TYPE": "PREROLL"},
        )
        ads1b = TagDateRangeAd(
            start=DATETIME_BASE,
            duration=1,
        )
        ads2 = TagDateRangeAd(
            start=DATETIME_BASE + timedelta(seconds=4),
            duration=4,
            custom={
                "X-TV-TWITCH-AD-ROLL-TYPE": "MIDROLL",
                "X-TV-TWITCH-AD-COMMERCIAL-ID": "123",
            },
        )
        ads3 = TagDateRangeAd(
            start=DATETIME_BASE + timedelta(seconds=8),
            duration=1,
            custom={
                "X-TV-TWITCH-AD-ROLL-TYPE": "MIDROLL",
                "X-TV-TWITCH-AD-COMMERCIAL-ID": "456",
                "X-TV-TWITCH-AD-POD-FILLED-DURATION": ".9",
            },
        )
        segments = self.subject(
            [
                Playlist(0, [ads1a, ads1b, Segment(0)]),
                Playlist(1, [ads1a, ads1b, Segment(1)]),
                Playlist(2, [Segment(2), Segment(3)]),
                Playlist(4, [ads2, Segment(4), Segment(5)]),
                Playlist(6, [ads2, Segment(6), Segment(7)]),
                Playlist(8, [ads3, Segment(8), Segment(9)], end=True),
            ],
            streamoptions={"low_latency": False},
        )

        self.await_write(10)
        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: s.num not in (0, 1, 4, 5, 6, 7, 8)), "Filters out all ad segments"
        assert all(self.called(s) for s in segments.values()), "Downloads all segments"
        assert self.thread.reader.worker.duration == pytest.approx(3.0), "Ad segments don't affect the output duration"
        assert mock_log.info.mock_calls == [
            call("Will skip ad segments"),
            call("Waiting for pre-roll ads to finish, be patient"),
            call("Detected advertisement break of 2 seconds"),
            call("Detected advertisement break of 4 seconds"),
            call("Detected advertisement break of 1 second"),
        ]

    @patch("streamlink.plugins.twitch.log")
    def test_hls_low_latency_has_prefetch(self, mock_log):
        segments = self.subject(
            [
                Playlist(0, [Segment(0), Segment(1), Segment(2), Segment(3), SegmentPrefetch(4), SegmentPrefetch(5)]),
                Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7), SegmentPrefetch(8), SegmentPrefetch(9)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        assert self.session.options.get("hls-live-edge") == 2
        assert self.session.options.get("hls-segment-stream-data")

        self.await_write(6)
        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: s.num >= 4), "Skips first four segments due to reduced live-edge"
        assert not any(self.called(s) for s in segments.values() if s.num < 4), "Doesn't download old segments"
        assert all(self.called(s) for s in segments.values() if s.num >= 4), "Downloads all remaining segments"
        assert mock_log.info.mock_calls == [
            call("Will skip ad segments"),
            call("Low latency streaming (HLS live edge: 2)"),
        ]

    @patch("streamlink.plugins.twitch.log")
    def test_hls_no_low_latency_has_prefetch(self, mock_log):
        segments = self.subject(
            [
                Playlist(0, [Segment(0), Segment(1), Segment(2), Segment(3), SegmentPrefetch(4), SegmentPrefetch(5)]),
                Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7), SegmentPrefetch(8), SegmentPrefetch(9)], end=True),
            ],
            streamoptions={"low_latency": False},
        )

        assert self.session.options.get("hls-live-edge") == 4
        assert not self.session.options.get("hls-segment-stream-data")

        self.await_write(8)
        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: s.num < 8), "Ignores prefetch segments"
        assert all(self.called(s) for s in segments.values() if s.num <= 7), "Ignores prefetch segments"
        assert not any(self.called(s) for s in segments.values() if s.num > 7), "Ignores prefetch segments"
        assert mock_log.info.mock_calls == [
            call("Will skip ad segments"),
        ]
        assert self.thread.reader.worker._reload_time == pytest.approx(3.0)

    @patch("streamlink.plugins.twitch.log")
    def test_hls_low_latency_no_prefetch(self, mock_log):
        self.subject(
            [
                Playlist(0, [Segment(0), Segment(1), Segment(2), Segment(3)]),
                Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        assert self.stream.low_latency

        self.await_write(6)
        self.await_read(read_all=True)
        assert mock_log.info.mock_calls == [
            call("Will skip ad segments"),
            call("Low latency streaming (HLS live edge: 2)"),
            call("This is not a low latency stream"),
        ]

    @patch("streamlink.plugins.twitch.log")
    def test_hls_low_latency_has_prefetch_has_preroll(self, mock_log):
        daterange = TagDateRangeAd(
            duration=4,
            custom={"X-TV-TWITCH-AD-ROLL-TYPE": "PREROLL"},
        )
        self.subject(
            [
                Playlist(0, [daterange, Segment(0), Segment(1), Segment(2), Segment(3)]),
                Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7), SegmentPrefetch(8), SegmentPrefetch(9)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        self.await_write(8)
        self.await_read(read_all=True)
        assert mock_log.info.mock_calls == [
            call("Will skip ad segments"),
            call("Low latency streaming (HLS live edge: 2)"),
            call("Waiting for pre-roll ads to finish, be patient"),
            call("Detected advertisement break of 4 seconds"),
        ]

    @patch("streamlink.plugins.twitch.log")
    def test_hls_low_latency_has_prefetch_no_preroll_with_prefetch_ads(self, mock_log):
        # segment 1 has a shorter duration, to mess with the extrapolation of the prefetch start times
        # segments 3-6 are ads
        Seg, Pre = Segment, SegmentPrefetch
        ads = [
            Tag("EXT-X-DISCONTINUITY"),
            TagDateRangeAd(
                start=DATETIME_BASE + timedelta(seconds=3),
                duration=4,
                custom={"X-TV-TWITCH-AD-ROLL-TYPE": "MIDROLL"},
            ),
        ]
        tls = Tag("EXT-X-TWITCH-LIVE-SEQUENCE", 7)
        # noinspection PyTypeChecker
        segments = self.subject(
            [
                # regular stream data with prefetch segments
                Playlist(0, [Seg(0), Seg(1, duration=0.5), Pre(2), Pre(3)]),
                # three prefetch segments, one regular (2) and two ads (3 and 4)
                Playlist(1, [Seg(1, duration=0.5), Pre(2), *ads, Pre(3), Pre(4)]),
                # all prefetch segments are gone once regular prefetch segments have shifted
                Playlist(2, [Seg(2, duration=1.5), *ads, Seg(3), Seg(4), Seg(5)]),
                # still no prefetch segments while ads are playing
                Playlist(3, [*ads, Seg(3), Seg(4), Seg(5), Seg(6)]),
                # new prefetch segments on the first regular segment occurrence
                Playlist(4, [*ads, Seg(4), Seg(5), Seg(6), tls, Seg(7), Pre(8), Pre(9)]),
                Playlist(5, [*ads, Seg(5), Seg(6), tls, Seg(7), Seg(8), Pre(9), Pre(10)]),
                Playlist(6, [*ads, Seg(6), tls, Seg(7), Seg(8), Seg(9), Pre(10), Pre(11)]),
                Playlist(7, [Seg(7), Seg(8), Seg(9), Seg(10), Pre(11), Pre(12)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        self.await_write(11)
        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: 2 <= s.num <= 3 or 7 <= s.num)
        assert mock_log.info.mock_calls == [
            call("Will skip ad segments"),
            call("Low latency streaming (HLS live edge: 2)"),
            call("Detected advertisement break of 4 seconds"),
        ]

    @patch("streamlink.plugins.twitch.log")
    def test_hls_low_latency_no_prefetch_has_preroll(self, mock_log):
        daterange = TagDateRangeAd(
            duration=4,
            custom={"X-TV-TWITCH-AD-ROLL-TYPE": "PREROLL"},
        )
        self.subject(
            [
                Playlist(0, [daterange, Segment(0), Segment(1), Segment(2), Segment(3)]),
                Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        self.await_write(6)
        self.await_read(read_all=True)
        assert mock_log.info.mock_calls == [
            call("Will skip ad segments"),
            call("Low latency streaming (HLS live edge: 2)"),
            call("Waiting for pre-roll ads to finish, be patient"),
            call("Detected advertisement break of 4 seconds"),
            call("This is not a low latency stream"),
        ]

    def test_hls_low_latency_no_ads_reload_time(self):
        Seg, SegPre = Segment, SegmentPrefetch
        self.subject(
            [
                Playlist(0, [Seg(0, duration=5), Seg(1, duration=7), Seg(2, duration=11), SegPre(3)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        self.await_write(4)
        self.await_read(read_all=True)
        assert self.thread.reader.worker._reload_time == pytest.approx(23 / 3)

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_prefetch_after_discontinuity(self, mock_log):
        segments = self.subject(
            [
                Playlist(0, [Segment(0), Segment(1)]),
                Playlist(2, [Segment(2), Segment(3), Tag("EXT-X-DISCONTINUITY"), SegmentPrefetch(4), SegmentPrefetch(5)]),
                Playlist(6, [Segment(6), Segment(7)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        self.await_write(8)
        assert self.await_read(read_all=True) == self.content(segments, cond=lambda seg: seg.num not in (4, 5))
        assert mock_log.warning.mock_calls == [
            call("Encountered a stream discontinuity. This is unsupported and will result in incoherent output data."),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_ignored_discontinuity(self, mock_log):
        Seg, Pre = Segment, SegmentPrefetch
        discontinuity = Tag("EXT-X-DISCONTINUITY")
        tls = Tag("EXT-X-TWITCH-LIVE-SEQUENCE", 1234)  # value is irrelevant
        segments = self.subject(
            [
                Playlist(0, [Seg(0), discontinuity, Seg(1)]),
                Playlist(2, [Seg(2), Seg(3), discontinuity, Seg(4), Seg(5)]),
                Playlist(6, [Seg(6), Seg(7), discontinuity, tls, Pre(8), Pre(9)]),
                Playlist(10, [Seg(10), Seg(11), discontinuity, tls, Pre(12), discontinuity, tls, Pre(13)], end=True),
            ],
            streamoptions={"low_latency": True},
        )

        self.await_write(14)
        assert self.await_read(read_all=True) == self.content(segments)
        assert mock_log.warning.mock_calls == []


class TestUsherService:
    @pytest.fixture(autouse=True)
    def caplog(self, caplog: pytest.LogCaptureFixture):
        caplog.set_level(1, "streamlink.plugins.twitch")
        return caplog

    @pytest.fixture()
    def plugin(self, request: pytest.FixtureRequest, session: Streamlink):
        return Twitch(
            session,
            "https://twitch.tv/twitch",
            options=Options(getattr(request, "param", {})),
        )

    @pytest.fixture()
    def endpoint(self, request: pytest.FixtureRequest, caplog: pytest.LogCaptureFixture, plugin: Twitch):
        param = getattr(request, "param", {})
        service = param.get("service", "channel")
        args = param.get("args", ("twitch",))

        token = {
            "expires": 9876543210,
            "channel": "twitch",
            "channel_id": 123,
            "user_id": 456,
            "user_ip": "127.0.0.1",
            "adblock": False,
            "geoblock_reason": "",
            "hide_ads": False,
            "server_ads": True,
            "show_ads": True,
        }

        return getattr(plugin.usher, service)(*args, token=json.dumps(token), sig="tokensignature")

    @pytest.mark.parametrize(
        ("endpoint", "expected_path", "logs"),
        [
            pytest.param(
                {"service": "channel", "args": ("TWITCH",)},
                "/api/v2/channel/hls/twitch.m3u8",
                [
                    (
                        "streamlink.plugins.twitch",
                        "debug",
                        "{'adblock': False, 'geoblock_reason': '', 'hide_ads': False, 'server_ads': True, 'show_ads': True}",
                    ),
                ],
                id="channel",
            ),
            pytest.param(
                {"service": "video", "args": ("1234567890",)},
                "/vod/v2/1234567890.m3u8",
                [],
                id="video",
            ),
        ],
        indirect=["endpoint"],
    )
    def test_service(self, caplog: pytest.LogCaptureFixture, endpoint: str, expected_path: str, logs: list):
        url = urlparse(endpoint)
        assert url.path == expected_path

        qs = dict(parse_qsl(url.query))
        assert qs.get("token")
        assert qs.get("sig")

        assert [(r.name, r.levelname, r.message) for r in caplog.get_records(when="setup")] == logs

    @pytest.mark.parametrize(
        "endpoint",
        [
            pytest.param({"service": "channel"}, id="channel"),
            pytest.param({"service": "video"}, id="video"),
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        ("plugin", "expected"),
        [
            pytest.param({}, "h264", id="unset"),
            pytest.param({"supported_codecs": []}, "h264", id="empty"),
            pytest.param({"supported_codecs": ["h264"]}, "h264", id="h264"),
            pytest.param({"supported_codecs": ["av1", "h264"]}, "av1,h264", id="av1,h264"),
            pytest.param({"supported_codecs": ["av1", "h264", "h265"]}, "av1,h264,h265", id="av1,h264,h265"),
        ],
        indirect=["plugin"],
    )
    def test_supported_codecs(self, plugin: Twitch, endpoint: str, expected: str):
        qs = dict(parse_qsl(urlparse(endpoint).query))
        assert qs.get("supported_codecs") == expected


class TestTwitchAPIAccessToken:
    @pytest.fixture(autouse=True)
    def _client_integrity_token(self, monkeypatch: pytest.MonkeyPatch):
        mock_client_integrity_token = Mock(return_value=("device-id", "client-integrity-token"))
        monkeypatch.setattr(Twitch, "_client_integrity_token", mock_client_integrity_token)

    @pytest.fixture()
    def plugin(self, request: pytest.FixtureRequest, session: Streamlink):
        options = Options(getattr(request, "param", {}))

        return Twitch(session, "https://twitch.tv/channelname", options)

    @pytest.fixture()
    def mock(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker):
        mock = requests_mock.post("https://gql.twitch.tv/gql", **getattr(request, "param", {"json": {}}))
        yield mock
        assert mock.call_count > 0
        payload = mock.last_request.json()  # type: ignore[union-attr, ty:unresolved-attribute]
        assert tuple(sorted(payload.keys())) == ("extensions", "operationName", "variables")
        assert payload.get("operationName") == "PlaybackAccessToken"
        assert payload.get("extensions") == {
            "persistedQuery": {
                "sha256Hash": "ed230aa1e33e07eebb8928504583da78a5173989fadfb1ac94be06a04f3cdbe9",
                "version": 1,
            },
        }

    @pytest.fixture()
    def _assert_live(self, mock: rm.Mocker):
        yield
        assert mock.last_request.json().get("variables") == {  # type: ignore[union-attr, ty:unresolved-attribute]
            "isLive": True,
            "isVod": False,
            "login": "channelname",
            "vodID": "",
            "playerType": "embed",
            "platform": "site",
        }

    @pytest.fixture()
    def _assert_vod(self, mock: rm.Mocker):
        yield
        assert mock.last_request.json().get("variables") == {  # type: ignore[union-attr, ty:unresolved-attribute]
            "isLive": False,
            "isVod": True,
            "login": "",
            "vodID": "vodid",
            "playerType": "embed",
            "platform": "site",
        }

    @pytest.mark.parametrize(
        ("plugin", "exp_headers", "exp_variables"),
        [
            (
                {},
                {"Client-ID": TwitchAPI.CLIENT_ID},
                {
                    "isLive": True,
                    "isVod": False,
                    "login": "channelname",
                    "vodID": "",
                    "playerType": "embed",
                    "platform": "site",
                },
            ),
            (
                {
                    "api-header": [
                        ("Authorization", "invalid data"),
                        ("Authorization", "OAuth 0123456789abcdefghijklmnopqrst"),
                    ],
                    "access-token-param": [
                        ("specialVariable", "specialValue"),
                        ("playerType", "frontpage"),
                        ("platform", "other"),
                    ],
                },
                {
                    "Client-ID": TwitchAPI.CLIENT_ID,
                    "Authorization": "OAuth 0123456789abcdefghijklmnopqrst",
                },
                {
                    "isLive": True,
                    "isVod": False,
                    "login": "channelname",
                    "vodID": "",
                    "playerType": "frontpage",
                    "platform": "other",
                    "specialVariable": "specialValue",
                },
            ),
        ],
        indirect=["plugin"],
    )
    def test_plugin_options(self, plugin: Twitch, mock: rm.Mocker, exp_headers: dict, exp_variables: dict):
        with pytest.raises(PluginError):
            plugin._access_token(True, "channelname")
        requestheaders = dict(mock.last_request._request.headers)  # type: ignore[union-attr, ty:unresolved-attribute]
        for header in plugin.session.http.headers.keys():
            del requestheaders[header]
        del requestheaders["Content-Type"]
        del requestheaders["Content-Length"]
        assert requestheaders == exp_headers
        assert mock.last_request.json().get("variables") == exp_variables  # type: ignore[union-attr, ty:unresolved-attribute]

    @pytest.mark.parametrize(
        ("session", "mock"),
        [
            pytest.param(
                {},
                {"json": {"data": {"streamPlaybackAccessToken": {"value": '{"channel":"foo"}', "signature": "sig"}}}},
                id="no-custom-user-agent",
            ),
            pytest.param(
                {"http-headers": {"User-Agent": "foo"}},
                {"json": {"data": {"streamPlaybackAccessToken": {"value": '{"channel":"foo"}', "signature": "sig"}}}},
                id="custom-user-agent",
            ),
        ],
        indirect=True,
    )
    def test_user_agent(self, plugin: Twitch, mock: rm.Mocker):
        plugin._access_token(True, "channelname")
        assert len(mock.request_history) > 0
        assert mock.request_history[0]._request.headers["User-Agent"] == useragents.DEFAULT

    @pytest.mark.usefixtures("_assert_live")
    @pytest.mark.parametrize(
        ("plugin", "mock"),
        [
            pytest.param(
                {
                    "force-client-integrity": False,
                },
                {
                    "json": {"data": {"streamPlaybackAccessToken": {"value": '{"channel":"foo"}', "signature": "sig"}}},
                },
                id="no-force-client-integrity",
            ),
            pytest.param(
                {
                    "force-client-integrity": True,
                },
                {
                    "json": {"data": {"streamPlaybackAccessToken": {"value": '{"channel":"foo"}', "signature": "sig"}}},
                },
                id="force-client-integrity",
            ),
        ],
        indirect=True,
    )
    def test_live_success(self, plugin: Twitch, mock: rm.Mocker):
        data = plugin._access_token(True, "channelname")
        assert data == ("sig", '{"channel":"foo"}', [])
        assert len(mock.request_history) == 1

    @pytest.mark.usefixtures("_assert_live")
    @pytest.mark.parametrize(
        "mock",
        [
            {
                "json": {"data": {"streamPlaybackAccessToken": None}},
            },
        ],
        indirect=True,
    )
    def test_live_failure(self, plugin: Twitch, mock: rm.Mocker):
        with pytest.raises(NoStreamsError):
            plugin._access_token(True, "channelname")
        assert len(mock.request_history) == 1, "Only gets the access token once when the channel is offline"

    @pytest.mark.usefixtures("_assert_vod")
    @pytest.mark.parametrize(
        "mock",
        [
            {
                "json": {"data": {"videoPlaybackAccessToken": {"value": '{"channel":"foo"}', "signature": "sig"}}},
            },
        ],
        indirect=True,
    )
    def test_vod_success(self, plugin: Twitch, mock: rm.Mocker):
        data = plugin._access_token(False, "vodid")
        assert data == ("sig", '{"channel":"foo"}', [])

    @pytest.mark.usefixtures("_assert_vod")
    @pytest.mark.parametrize(
        "mock",
        [
            {
                "json": {"data": {"videoPlaybackAccessToken": None}},
            },
        ],
        indirect=True,
    )
    def test_vod_failure(self, plugin: Twitch, mock: rm.Mocker):
        with pytest.raises(NoStreamsError):
            plugin._access_token(False, "vodid")
        assert len(mock.request_history) == 1, "Only gets the access token once when the VOD doesn't exist"

    @pytest.mark.usefixtures("_assert_live")
    @pytest.mark.parametrize(
        ("plugin", "mock"),
        [
            (
                {
                    "api-header": [("Authorization", "OAuth invalid-token")],
                },
                {
                    "status_code": 401,
                    "json": {"error": "Unauthorized", "status": 401, "message": 'The "Authorization" token is invalid.'},
                },
            ),
        ],
        indirect=True,
    )
    def test_auth_failure(self, plugin: Twitch, mock: rm.Mocker):
        with pytest.raises(PluginError, match=r'^Unauthorized: The "Authorization" token is invalid\.$'):
            plugin._access_token(True, "channelname")
        assert len(mock.request_history) == 2, "Always tries again on error, with integrity-token on second attempt"

        headers: dict = mock.request_history[0]._request.headers
        assert headers["Authorization"] == "OAuth invalid-token"
        assert "Device-Id" not in headers
        assert "Client-Integrity" not in headers

        headers = mock.request_history[1]._request.headers
        assert headers["Authorization"] == "OAuth invalid-token"
        assert headers["Device-Id"] == "device-id"
        assert headers["Client-Integrity"] == "client-integrity-token"

    @pytest.mark.usefixtures("_assert_live")
    @pytest.mark.parametrize(
        ("plugin", "mock"),
        [
            (
                {
                    "force-client-integrity": False,
                    "api-header": [("Authorization", "OAuth invalid-token")],
                },
                {
                    "response_list": [
                        {
                            "status_code": 401,
                            "json": {"errors": [{"message": "failed integrity check"}]},
                        },
                        {
                            "json": {"data": {"streamPlaybackAccessToken": {"value": '{"channel":"foo"}', "signature": "sig"}}},
                        },
                    ],
                },
            ),
        ],
        indirect=True,
    )
    def test_integrity_check_not_forced(self, plugin: Twitch, mock: rm.Mocker):
        data = plugin._access_token(True, "channelname")
        assert data == ("sig", '{"channel":"foo"}', [])
        assert len(mock.request_history) == 2, "Always tries again on error, with integrity-token on second attempt"

        headers: dict = mock.request_history[0]._request.headers
        assert headers["Authorization"] == "OAuth invalid-token"
        assert "Device-Id" not in headers
        assert "Client-Integrity" not in headers

        headers = mock.request_history[1]._request.headers
        assert headers["Authorization"] == "OAuth invalid-token"
        assert headers["Device-Id"] == "device-id"
        assert headers["Client-Integrity"] == "client-integrity-token"

    @pytest.mark.usefixtures("_assert_live")
    @pytest.mark.parametrize(
        ("plugin", "mock"),
        [
            (
                {
                    "force-client-integrity": True,
                    "api-header": [("Authorization", "OAuth invalid-token")],
                },
                {
                    "response_list": [
                        {
                            "json": {"data": {"streamPlaybackAccessToken": {"value": '{"channel":"foo"}', "signature": "sig"}}},
                        },
                    ],
                },
            ),
        ],
        indirect=True,
    )
    def test_integrity_check_forced(self, plugin: Twitch, mock: rm.Mocker):
        data = plugin._access_token(True, "channelname")
        assert data == ("sig", '{"channel":"foo"}', [])
        assert len(mock.request_history) == 1

        headers: dict = mock.request_history[0]._request.headers
        assert headers["Authorization"] == "OAuth invalid-token"
        assert headers["Device-Id"] == "device-id"
        assert headers["Client-Integrity"] == "client-integrity-token"

    @pytest.mark.usefixtures("_assert_vod")
    @pytest.mark.parametrize(
        ("plugin", "mock"),
        [
            (
                {
                    "force-client-integrity": True,
                    "api-header": [("Authorization", "OAuth invalid-token")],
                },
                {
                    "response_list": [
                        {
                            "json": {"data": {"videoPlaybackAccessToken": {"value": '{"channel":"foo"}', "signature": "sig"}}},
                        },
                    ],
                },
            ),
        ],
        indirect=True,
    )
    def test_integrity_check_vod(self, plugin: Twitch, mock: rm.Mocker):
        data = plugin._access_token(False, "vodid")
        assert data == ("sig", '{"channel":"foo"}', [])
        assert len(mock.request_history) == 1

        headers: dict = mock.request_history[0]._request.headers
        assert headers["Authorization"] == "OAuth invalid-token"
        assert "Device-Id" not in headers
        assert "Client-Integrity" not in headers


class TestTwitchHLSMultivariantResponse:
    @pytest.fixture()
    def plugin(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker, session: Streamlink):
        requests_mock.get("mock://multivariant", **getattr(request, "param", {}))
        return Twitch(session, "https://twitch.tv/channelname")

    @pytest.mark.parametrize(
        ("plugin", "streamid", "raises", "streams", "log"),
        [
            pytest.param(
                {"text": "#EXTM3U\n"},
                "123",
                nullcontext(),
                {},
                [],
                id="success",
            ),
            pytest.param(
                {"text": "Not an HLS playlist"},
                "123",
                pytest.raises(PluginError),
                {},
                [],
                id="invalid HLS playlist",
            ),
            pytest.param(
                {
                    "status_code": 404,
                    "json": [
                        {
                            "url": "mock://multivariant",
                            "error": "twirp error not_found: transcode does not exist",
                            "error_code": "transcode_does_not_exist",
                            "type": "error",
                        },
                    ],
                },
                None,
                nullcontext(),
                None,
                [],
                id="offline",
            ),
            pytest.param(
                {
                    "status_code": 403,
                    "json": [
                        {
                            "url": "mock://multivariant",
                            "error": "Content Restricted In Region",
                            "error_code": "content_geoblocked",
                            "type": "error",
                        },
                    ],
                },
                "123",
                nullcontext(),
                None,
                [("streamlink.plugins.twitch", "error", "Content Restricted In Region")],
                id="geo restriction",
            ),
            pytest.param(
                {
                    "status_code": 404,
                    "text": "Not found",
                },
                "123",
                nullcontext(),
                None,
                [],
                id="non-json error response",
            ),
        ],
        indirect=["plugin"],
    )
    def test_multivariant_response(
        self,
        monkeypatch: pytest.MonkeyPatch,
        caplog: pytest.LogCaptureFixture,
        plugin: Twitch,
        streamid: str | None,
        raises: nullcontext,
        streams: dict | None,
        log: list,
    ):
        caplog.set_level("error", "streamlink.plugins.twitch")
        monkeypatch.setattr(plugin, "get_id", Mock(return_value=streamid))
        with raises:
            assert plugin._get_hls_streams("mock://multivariant", []) == streams
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == log


class TestTwitchMetadata:
    @pytest.fixture()
    def metadata(self, request: pytest.FixtureRequest, session: Streamlink):
        url = getattr(request, "param", "")
        plugin = Twitch(session, url)

        return plugin.get_id(), plugin.get_author(), plugin.get_category(), plugin.get_title()

    @pytest.fixture()
    def mock_request_channel(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker):
        data = getattr(request, "param", True)

        return requests_mock.post(
            "https://gql.twitch.tv/gql",
            json=[
                {
                    "data": {
                        "userOrError": {"userDoesNotExist": "error"}
                        if not data
                        else {
                            "displayName": "channel name",
                        },
                    },
                },
                {
                    "data": {
                        "user": None
                        if not data
                        else {
                            "lastBroadcast": {
                                "title": "channel status",
                            },
                            "stream": {
                                "id": "stream id",
                                "game": {
                                    "name": "channel game",
                                },
                            },
                        },
                    },
                },
            ],
        )

    @pytest.fixture()
    def mock_request_video(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker):
        data = getattr(request, "param", True)

        return requests_mock.post(
            "https://gql.twitch.tv/gql",
            json={
                "data": {
                    "video": None
                    if not data
                    else {
                        "id": "video id",
                        "title": "video title",
                        "game": {
                            "displayName": "video game",
                        },
                        "owner": {
                            "displayName": "channel name",
                        },
                    },
                },
            },
        )

    @pytest.fixture()
    def mock_request_clip(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker):
        data = getattr(request, "param", True)

        return requests_mock.post(
            "https://gql.twitch.tv/gql",
            json={
                "data": {
                    "clip": None
                    if not data
                    else {
                        "id": "clip id",
                        "broadcaster": {
                            "displayName": "channel name",
                        },
                        "game": {
                            "name": "game name",
                        },
                        "title": "clip title",
                    },
                },
            },
        )

    @pytest.mark.parametrize(("mock_request_channel", "metadata"), [(True, "https://twitch.tv/foo")], indirect=True)
    def test_metadata_channel(self, mock_request_channel, metadata):
        assert metadata == ("stream id", "channel name", "channel game", "channel status")
        assert mock_request_channel.call_count == 1
        assert mock_request_channel.request_history[0].json() == [
            {
                "operationName": "ChannelShell",
                "extensions": {
                    "persistedQuery": {
                        "version": 1,
                        "sha256Hash": "fea4573a7bf2644f5b3f2cbbdcbee0d17312e48d2e55f080589d053aad353f11",
                    },
                },
                "variables": {
                    "login": "foo",
                },
            },
            {
                "operationName": "StreamMetadata",
                "extensions": {
                    "persistedQuery": {
                        "version": 1,
                        "sha256Hash": "b57f9b910f8cd1a4659d894fe7550ccc81ec9052c01e438b290fd66a040b9b93",
                    },
                },
                "variables": {
                    "channelLogin": "foo",
                    "includeIsDJ": True,
                },
            },
        ]

    @pytest.mark.parametrize(("mock_request_channel", "metadata"), [(False, "https://twitch.tv/foo")], indirect=True)
    def test_metadata_channel_no_data(self, mock_request_channel, metadata):
        assert metadata == (None, None, None, None)
        assert mock_request_channel.call_count == 1

    @pytest.mark.parametrize(("mock_request_video", "metadata"), [(True, "https://twitch.tv/videos/1337")], indirect=True)
    def test_metadata_video(self, mock_request_video, metadata):
        assert metadata == ("video id", "channel name", "video game", "video title")
        assert mock_request_video.call_count == 1
        assert mock_request_video.request_history[0].json() == {
            "operationName": "VideoMetadata",
            "extensions": {
                "persistedQuery": {
                    "version": 1,
                    "sha256Hash": "45111672eea2e507f8ba44d101a61862f9c56b11dee09a15634cb75cb9b9084d",
                },
            },
            "variables": {
                "channelLogin": "",
                "videoID": "1337",
            },
        }

    @pytest.mark.parametrize(("mock_request_video", "metadata"), [(False, "https://twitch.tv/videos/1337")], indirect=True)
    def test_metadata_video_no_data(self, mock_request_video, metadata):
        assert metadata == (None, None, None, None)
        assert mock_request_video.call_count == 1

    @pytest.mark.parametrize(("mock_request_clip", "metadata"), [(True, "https://clips.twitch.tv/foo")], indirect=True)
    def test_metadata_clip(self, mock_request_clip, metadata):
        assert metadata == ("clip id", "channel name", "game name", "clip title")
        assert mock_request_clip.call_count == 1
        assert mock_request_clip.request_history[0].json() == {
            "operationName": "ShareClipRenderStatus",
            "extensions": {
                "persistedQuery": {
                    "version": 1,
                    "sha256Hash": "1844261bb449fa51e6167040311da4a7a5f1c34fe71c71a3e0c4f551bc30c698",
                },
            },
            "variables": {
                "slug": "foo",
            },
        }

    @pytest.mark.parametrize(("mock_request_clip", "metadata"), [(False, "https://clips.twitch.tv/foo")], indirect=True)
    def test_metadata_clip_no_data(self, mock_request_clip, metadata):
        assert metadata == (None, None, None, None)



from streamlink.plugins.nasaplus import NASAPlus
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlNASAPlus(PluginCanHandleUrl):
    __plugin__ = NASAPlus

    should_match = [
        # live / rebroadcast
        "https://plus.nasa.gov/scheduled-video/iss-expedition-70-in-flight-educational-event-with-the-creative-learning-academy-in-pensacola-florida-and-nasa-flight-engineer-jasmin-moghbeli/",
        # VOD
        "https://plus.nasa.gov/video/moon-101-introduction-to-the-moon/",
    ]



import pytest
from setuptools.command.egg_info import egg_info

from build_backend import _filter_cmd_option_args


@pytest.mark.parametrize(
    ("config_settings", "expected", "options"),
    [
        pytest.param(
            None,
            None,
            egg_info.user_options,
            id="Empty config_settings",
        ),
        pytest.param(
            {"foo": "bar"},
            {"foo": "bar"},
            egg_info.user_options,
            id="No --build-option key",
        ),
        pytest.param(
            {"--build-option": "--egg-base=foo/bar -e baz/qux --tag-build foo -b bar --tag-date --no-date -D"},
            {"--build-option": "--egg-base=foo/bar -e baz/qux --tag-build foo -b bar --tag-date --no-date -D"},
            egg_info.user_options,
            id="All egg_info options",
        ),
        pytest.param(
            {"--build-option": "--foo --bar --baz"},
            {},
            egg_info.user_options,
            id="Options unknown to egg_info",
        ),
        pytest.param(
            {"--build-option": "-p win32 --plat-name win32 --plat-name=win32"},
            {},
            egg_info.user_options,
            id="bdist_wheel --plat-name option",
        ),
    ],
)
def test_filter_cmd_option_args(config_settings: dict, expected: str, options: list):
    _filter_cmd_option_args(config_settings, "--build-option", options)
    assert config_settings == expected



from unittest.mock import Mock, patch

import pytest

import streamlink.utils.l10n as l10n


class TestLocalization:
    @pytest.mark.parametrize(
        ("locale", "expected"),
        [
            ("en_US", "en_US"),
            ("ko_KR", "ko_KR"),
        ],
    )
    def test_valid(self, locale, expected):
        locale = l10n.Localization(locale)
        assert locale.language_code == expected

    @pytest.mark.parametrize(
        "locale",
        [
            "enUS",
            "eng_US",
            "en_USA",
        ],
    )
    def test_invalid(self, locale):
        with pytest.raises(LookupError):
            l10n.Localization(locale)

    @pytest.mark.parametrize(
        "mock_getlocale",
        [
            Mock(return_value=(None, None)),
            Mock(return_value=("en_150", None)),
            Mock(side_effect=ValueError("unknown locale: foo_bar")),
        ],
    )
    def test_default(self, mock_getlocale: Mock):
        with patch("locale.getlocale", mock_getlocale):
            locale = l10n.Localization()
            assert locale.language_code == "en_US"
            assert locale.equivalent(language="en", country="US")

    def test_setter(self):
        with patch("locale.getlocale", Mock(return_value=(None, None))):
            locale = l10n.Localization()
            assert locale.language_code == "en_US"
            assert locale.equivalent(language="en", country="US")

            locale.language_code = "de_DE"
            assert locale.language_code == "de_DE"
            assert locale.equivalent(language="de", country="DE")


class TestLocalizationEquality:
    @pytest.mark.parametrize(
        ("language", "country"),
        [
            (None, None),
            ("eng", None),
            ("en", None),
            ("en", "CA"),
            ("en", "CAN"),
            ("en", "Canada"),
        ],
    )
    def test_equivalent(self, language, country):
        locale = l10n.Localization("en_CA")
        assert locale.equivalent(language, country)

    @pytest.mark.parametrize(
        ("language", "country"),
        [
            ("en", "ca"),
            ("English", "Canada"),
        ],
    )
    def test_equivalent_language_country_objects(self, language, country):
        locale = l10n.Localization("en_CA")
        assert locale.equivalent(l10n.Language.get(language), l10n.Country.get(country))

    @pytest.mark.parametrize(
        "language",
        [
            "fra",
            "fre",
        ],
    )
    def test_equivalent_remap(self, language):
        locale = l10n.Localization("fr_FR")
        assert locale.equivalent(language)

    @pytest.mark.parametrize(
        ("language", "country"),
        [
            ("eng", None),
            ("en", None),
            ("en", "US"),
            ("en", "Canada"),
            ("en", "ES"),
            ("en", "Spain"),
            ("en", "UNKNOWN"),
            ("UNKNOWN", "Spain"),
        ],
    )
    def test_not_equivalent(self, language, country):
        locale = l10n.Localization("es_ES")
        assert not locale.equivalent(language, country)


class TestCountry:
    @pytest.mark.parametrize(
        ("country", "attr", "expected"),
        [
            ("USA", "alpha2", "US"),
            ("GB", "alpha2", "GB"),
            ("Canada", "name", "Canada"),
        ],
    )
    def test_get_country(self, country, attr, expected):
        assert getattr(l10n.Localization.get_country(country), attr) == expected

    @pytest.mark.parametrize(
        "country",
        [
            "XE",
            "XEX",
            "Nowhere",
        ],
    )
    def test_get_country_miss(self, country):
        with pytest.raises(LookupError):
            l10n.Localization.get_country(country)

    def test_hash(self):
        one = l10n.Country("DE", "DEU", "276", "Germany", "Federal Republic of Germany")
        two = l10n.Country("DE", "DEU", "276", "Germany", "Federal Republic of Germany")
        mapping = {one: "1", two: "2"}
        assert one is not two
        assert mapping[one] is mapping[two]

    def test_country_compare(self):
        assert l10n.Country("AA", "AAA", "001", "Test") == l10n.Country("AA", "AAA", "001", "Test")

    def test_country_str(self):
        assert (
            str(l10n.Localization.get_country("Germany"))
            == "Country('DE', 'DEU', '276', 'Germany', official_name='Federal Republic of Germany')"
        )


class TestLanguage:
    @pytest.mark.parametrize(
        ("language", "attr", "expected"),
        [
            ("en", "alpha3", "eng"),
            ("fra", "bibliographic", "fre"),
            ("fre", "alpha3", "fra"),
            ("gre", "bibliographic", "gre"),
        ],
    )
    def test_get_language(self, language, attr, expected):
        assert getattr(l10n.Localization.get_language(language), attr) == expected

    @pytest.mark.parametrize(
        "language",
        [
            "00",
            "000",
            "0000",
        ],
    )
    def test_get_language_miss(self, language):
        with pytest.raises(LookupError):
            l10n.Localization.get_language(language)

    def test_hash(self):
        one = l10n.Language("de", "deu", "German", "ger")
        two = l10n.Language("de", "deu", "German", "ger")
        mapping = {one: "1", two: "2"}
        assert one is not two
        assert mapping[one] is mapping[two]

    def test_language_compare(self):
        assert l10n.Language("AA", "AAA", "Test") == l10n.Language("AA", None, "Test")
        assert l10n.Language("BB", "BBB", "Test") != l10n.Language("AA", None, "Test")

    def test_language_str(self):
        assert str(l10n.Localization.get_language("German")) == "Language('de', 'deu', 'German', bibliographic='ger')"

    # issue #3517: language lookups without alpha2 but with alpha3 codes should not raise
    def test_language_a3_no_a2(self):
        lang = l10n.Localization.get_language("des")
        assert lang.alpha2 == ""
        assert lang.alpha3 == "des"
        assert lang.name == "Desano"
        assert lang.bibliographic == ""

    # issue #3057: generic "en" lookups via pycountry yield the "En" language, but not "English"
    @pytest.mark.parametrize(
        "language",
        [
            "en",
            "eng",
            "English",
        ],
    )
    def test_language_en(self, language):
        lang = l10n.Localization.get_language(language)
        assert lang.alpha2 == "en"
        assert lang.alpha3 == "eng"
        assert lang.name == "English"
        assert lang.bibliographic == ""



from argparse import Namespace
from unittest.mock import Mock

import pytest

from streamlink.plugin import Plugin
from streamlink_cli.main import get_formatter
from streamlink_cli.utils import datetime


@pytest.fixture(scope="module")
def plugin():
    class FakePlugin(Plugin):
        __module__ = "FAKE"

        def _get_streams(self):  # pragma: no cover
            pass

    plugin = FakePlugin(Mock(), "https://foo/bar")
    plugin.id = "ID"
    plugin.author = "AUTHOR"
    plugin.category = "CATEGORY"
    plugin.title = "TITLE"

    return plugin


@pytest.mark.parametrize(
    ("formatterinput", "expected"),
    [
        pytest.param("{url}", "https://foo/bar", id="url"),
        pytest.param("{plugin}", "FAKE", id="plugin-name"),
        pytest.param("{id}", "ID", id="id"),
        pytest.param("{author}", "AUTHOR", id="author"),
        pytest.param("{category}", "CATEGORY", id="category"),
        pytest.param("{game}", "CATEGORY", id="category-fallback-game"),
        pytest.param("{title}", "TITLE", id="title"),
        pytest.param("{time}", "2000-01-01_00-00-00", id="time"),
        pytest.param("{time:%Y}", "2000", id="time-formatted"),
    ],
)
def test_get_formatter(monkeypatch: pytest.MonkeyPatch, plugin: Plugin, formatterinput: str, expected: str):
    # workaround for freezegun not being able to patch the subclassed datetime class in streamlink_cli.utils
    # which defines the default datetime->str conversion format (needed for path outputs)
    monkeypatch.setattr("streamlink_cli.utils.datetime.now", Mock(return_value=datetime(2000, 1, 1, 0, 0, 0, 0)))
    monkeypatch.setattr("streamlink_cli.main.args", Namespace(url="https://foo/bar"))

    formatter = get_formatter(plugin)
    assert formatter.title(formatterinput) == expected



from streamlink.plugins.tf1 import TF1
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTF1(PluginCanHandleUrl):
    __plugin__ = TF1

    should_match_groups = [
        (("live", "https://tf1.fr/tf1/direct"), {"live": "tf1"}),
        (("live", "https://www.tf1.fr/tf1/direct"), {"live": "tf1"}),
        (("live", "https://www.tf1.fr/tfx/direct"), {"live": "tfx"}),
        (("live", "https://www.tf1.fr/tf1-series-films/direct"), {"live": "tf1-series-films"}),
        (("stream", "https://www.tf1.fr/chante-69061019/direct"), {"stream": "chante-69061019"}),
        (("stream", "https://www.tf1.fr/thriller-fiction-89242722/direct"), {"stream": "thriller-fiction-89242722"}),
        (("lci", "https://lci.fr/direct"), {}),
        (("lci", "https://www.lci.fr/direct"), {}),
        (("lci", "https://tf1info.fr/direct/"), {}),
        (("lci", "https://www.tf1info.fr/direct/"), {}),
    ]

    should_not_match = [
        "http://tf1.fr/direct",
        "http://www.tf1.fr/direct",
    ]



from streamlink.plugins.rtvs import Rtvs
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlRtvs(PluginCanHandleUrl):
    __plugin__ = Rtvs

    should_match = [
        "https://www.rtvs.sk/televizia/live-1",
        "https://www.rtvs.sk/televizia/live-2",
        "https://www.rtvs.sk/televizia/live-3",
        "https://www.rtvs.sk/televizia/live-o",
        "https://www.rtvs.sk/televizia/live-rtvs",
        "https://www.rtvs.sk/televizia/live-nr-sr",
        "https://www.rtvs.sk/televizia/sport",
    ]

    should_not_match = [
        "http://www.rtvs.sk/",
        "http://www.rtvs.sk/televizia/archiv",
        "http://www.rtvs.sk/televizia/program",
    ]



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock, call

import pytest

from streamlink_cli.utils.versioncheck import check_version, get_latest, log


if TYPE_CHECKING:
    import requests_mock as rm


@pytest.fixture(autouse=True)
def _logger(caplog: pytest.LogCaptureFixture):
    caplog.set_level(1, "streamlink.cli")


def test_logger_name():
    assert log.name == "streamlink.cli"


class TestGetLatest:
    @pytest.fixture()
    def pypi(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker):
        response = requests_mock.register_uri(
            "GET",
            "https://pypi.python.org/pypi/streamlink/json",
            **getattr(request, "param", {}),
        )
        yield response
        assert response.call_count == 1

    @pytest.mark.parametrize(
        ("pypi", "error"),
        [
            (
                {"status_code": 500},
                "Error while retrieving version data from PyPI API: "
                + "500 Server Error: None for url: https://pypi.python.org/pypi/streamlink/json",
            ),
            (
                {"text": "no JSON"},
                "Could not parse JSON data from PyPI API response",
            ),
            (
                {"json": {"foo": "bar"}},
                "Error while retrieving version data from PyPI API: 'NoneType' object has no attribute 'get'",
            ),
        ],
        indirect=["pypi"],
    )
    def test_request_error(self, caplog: pytest.LogCaptureFixture, pypi, error):
        assert not get_latest()
        assert [(record.levelname, str(record.message)) for record in caplog.records] == [("error", error)]

    @pytest.mark.parametrize("pypi", [{"json": {"info": {"version": "1.2.3"}}}], indirect=True)
    def test_request_success(self, caplog: pytest.LogCaptureFixture, pypi):
        assert get_latest() == "1.2.3"
        assert not caplog.records


class TestVersionCheck:
    @pytest.fixture(autouse=True)
    def _current(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("streamlink_cli.utils.versioncheck.streamlink_version", "1.0.0")

    @pytest.fixture()
    def latest(self, request, monkeypatch: pytest.MonkeyPatch):
        mock_get_latest = Mock(return_value=getattr(request, "param", "1.2.3"))
        monkeypatch.setattr("streamlink_cli.utils.versioncheck.get_latest", mock_get_latest)
        return mock_get_latest

    @pytest.fixture()
    def cache(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
        mock_cache = Mock()
        mock_cache.get.side_effect = getattr(request, "param", {}).get
        MockCache = Mock(return_value=mock_cache)
        monkeypatch.setattr("streamlink_cli.utils.versioncheck.Cache", MockCache)
        yield mock_cache
        assert MockCache.call_args_list == [call(filename="cli.json")]

    @pytest.mark.parametrize("cache", [{}], indirect=True)
    def test_auto_uncached_outdated(self, caplog: pytest.LogCaptureFixture, cache: Mock, latest: Mock):
        assert not check_version()
        assert latest.call_args_list == [call()]
        assert cache.set.call_args_list == [
            call("latest_version", "1.2.3", 86400),
            call("version_info_printed", True, 21600),
        ]
        assert [(record.levelname, str(record.message)) for record in caplog.records] == [
            ("info", "A new version of Streamlink (1.2.3) is available!"),
        ]

    @pytest.mark.parametrize(("cache", "latest"), [({}, "1.0.0")], indirect=True)
    def test_auto_uncached_uptodate(self, caplog: pytest.LogCaptureFixture, cache: Mock, latest: Mock):
        assert check_version()
        assert latest.call_args_list == [call()]
        assert cache.set.call_args_list == [
            call("latest_version", "1.0.0", 86400),
        ]
        assert not caplog.records

    @pytest.mark.parametrize("cache", [{"latest_version": "1.2.3", "version_info_printed": False}], indirect=True)
    def test_auto_cached_outdated(self, caplog: pytest.LogCaptureFixture, cache: Mock, latest: Mock):
        assert not check_version()
        assert not latest.call_args_list
        assert cache.set.call_args_list == [call("version_info_printed", True, 21600)]
        assert [(record.levelname, str(record.message)) for record in caplog.records] == [
            ("info", "A new version of Streamlink (1.2.3) is available!"),
        ]

    @pytest.mark.parametrize("cache", [{"latest_version": "1.2.3", "version_info_printed": True}], indirect=True)
    def test_auto_cached_printed(self, caplog: pytest.LogCaptureFixture, cache: Mock, latest: Mock):
        assert check_version()
        assert not latest.call_args_list
        assert not cache.set.call_args_list
        assert not caplog.records

    @pytest.mark.parametrize(
        "cache",
        [
            {},
            {"version_info_printed": True},
        ],
        indirect=True,
    )
    def test_forced_outdated(self, caplog: pytest.LogCaptureFixture, cache: Mock, latest: Mock):
        assert not check_version(True)
        assert latest.call_args_list == [call()]
        assert cache.set.call_args_list == [
            call("latest_version", "1.2.3", 86400),
            call("version_info_printed", True, 21600),
        ]
        assert [(record.levelname, str(record.message)) for record in caplog.records] == [
            ("info", "A new version of Streamlink (1.2.3) is available!"),
        ]

    @pytest.mark.parametrize(
        ("cache", "latest"),
        [
            ({}, "1.0.0"),
            ({"version_info_printed": True}, "1.0.0"),
        ],
        indirect=True,
    )
    def test_forced_uptodate(self, caplog: pytest.LogCaptureFixture, cache: Mock, latest: Mock):
        assert check_version(True)
        assert latest.call_args_list == [call()]
        assert cache.set.call_args_list == [
            call("latest_version", "1.0.0", 86400),
        ]
        assert [(record.levelname, str(record.message)) for record in caplog.records] == [
            ("info", "Your Streamlink version (1.0.0) is up to date!"),
        ]

    @pytest.mark.parametrize(("cache", "latest"), [({}, "")], indirect=True)
    def test_error_get_latest(self, caplog: pytest.LogCaptureFixture, cache: Mock, latest: Mock):
        assert not check_version(True)
        assert latest.call_args_list == [call()]
        assert not cache.set.call_args_list
        assert not caplog.records  # error gets logged by get_latest()

    @pytest.mark.parametrize(("cache", "latest"), [({}, "not a semver version string")], indirect=True)
    def test_error_get_latest_version(self, caplog: pytest.LogCaptureFixture, cache: Mock, latest: Mock):
        assert not check_version(True)
        assert latest.call_args_list == [call()]
        assert not cache.set.call_args_list
        assert [(record.levelname, str(record.message)) for record in caplog.records] == [
            ("error", f"Error while parsing version: Invalid version string: '{latest.return_value}'"),
        ]



from streamlink.plugins.nrk import NRK
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlNRK(PluginCanHandleUrl):
    __plugin__ = NRK

    should_match = [
        "https://tv.nrk.no/direkte/nrk1",
        "https://tv.nrk.no/direkte/nrk2",
        "https://tv.nrk.no/direkte/nrk3",
        "https://tv.nrk.no/direkte/nrksuper",
        "https://tv.nrk.no/serie/nytt-paa-nytt/2020/MUHH43003020",
        "https://tv.nrk.no/serie/kongelige-fotografer/sesong/1/episode/2/avspiller",
        "https://tv.nrk.no/program/NNFA51102617",
        "https://radio.nrk.no/direkte/p1",
        "https://radio.nrk.no/direkte/p2",
        "https://radio.nrk.no/podkast/oppdatert/l_5005d62a-7f4f-4581-85d6-2a7f4f2581f2",
    ]

    should_not_match = [
        "https://tv.nrk.no/",
        "https://radio.nrk.no/",
        "https://nrk.no/",
    ]



from streamlink.plugins.tv3cat import TV3Cat
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTV3Cat(PluginCanHandleUrl):
    __plugin__ = TV3Cat

    should_match_groups = [
        (("live", "https://www.3cat.cat/3cat/directes/tv3/"), {"ident": "tv3"}),
        (("live", "https://www.ccma.cat/3cat/directes/tv3/"), {"ident": "tv3"}),
        (("live", "https://www.ccma.cat/3cat/directes/324/"), {"ident": "324"}),
        (("live", "https://www.ccma.cat/3cat/directes/esport3/"), {"ident": "esport3"}),
        (("live", "https://www.ccma.cat/3cat/directes/sx3/"), {"ident": "sx3"}),
        (("live", "https://www.ccma.cat/3cat/directes/catalunya-radio/"), {"ident": "catalunya-radio"}),
        (
            ("vod", "https://www.3cat.cat/3cat/t1xc1-arribada/video/6260741/"),
            {"ident": "6260741"},
        ),
        (
            ("vod", "https://www.ccma.cat/3cat/t1xc1-arribada/video/6260741/"),
            {"ident": "6260741"},
        ),
        (
            ("vod", "https://www.ccma.cat/3cat/merli-els-peripatetics-capitol-1/video/5549976/"),
            {"ident": "5549976"},
        ),
        (
            ("vod", "https://www.ccma.cat/3cat/buscant-la-sostenibilitat-i-la-tecnologia-del-futur/video/6268863/"),
            {"ident": "6268863"},
        ),
    ]



from __future__ import annotations

import subprocess
from contextlib import nullcontext
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, call, patch

import pytest

from streamlink.exceptions import StreamlinkWarning
from streamlink_cli.output.player import (
    PlayerArgs,
    PlayerArgsMPV,
    PlayerArgsPotplayer,
    PlayerArgsVLC,
    PlayerOutput,
)


if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from contextlib import AbstractContextManager


class TestPlayerArgs:
    @pytest.fixture()
    def playerargs(self, request: pytest.FixtureRequest):
        return PlayerArgs(**getattr(request, "param", {}))

    @pytest.fixture()
    def _playerargv(self, request: pytest.FixtureRequest, playerargs: PlayerArgs):
        assert playerargs.build() == getattr(request, "param", None)

    # noinspection PyTestParametrized
    @pytest.mark.usefixtures("playerargs", "_playerargv")
    @pytest.mark.parametrize(
        ("playerargs", "_playerargv"),
        [
            pytest.param(
                dict(path=Path("player")),
                ["player", "-"],
                id="Without player args",
            ),
            pytest.param(
                dict(path=Path("path to player"), args='p1 "1 2 3" p2="4 5 6" \'p3=7 8 9\' "a=\'b c\' d" \'e"f g"h\''),
                ["path to player", "p1", "1 2 3", "p2=4 5 6", "p3=7 8 9", "a='b c' d", 'e"f g"h', "-"],
                id="Player args tokenization",
            ),
            pytest.param(
                dict(path=Path("player"), args="param1 param2=value"),
                ["player", "param1", "param2=value", "-"],
                id="Implicit playerinput variable",
            ),
            pytest.param(
                dict(path=Path("player"), args="{playerinput} param1 param2=value"),
                ["player", "-", "param1", "param2=value"],
                id="Explicit playerinput variable",
            ),
            pytest.param(
                dict(path=Path("player"), args="{playerinput} param1 param2=value {playerinput}"),
                ["player", "-", "param1", "param2=value", "-"],
                id="Duplicate playerinput variable",
            ),
            pytest.param(
                dict(path=Path("player"), args="param1 {playerinput}-stdin param2"),
                ["player", "param1", "--stdin", "param2"],
                id="Combination of playerinput variable",
            ),
            pytest.param(
                dict(path=Path("player"), args="param1 param2=value {unknown}"),
                ["player", "param1", "param2=value", "{unknown}", "-"],
                id="Unknown player args variable",
            ),
            pytest.param(
                dict(path=Path("/usr/bin/player")),
                ["/usr/bin/player", "-"],
                marks=pytest.mark.posix_only,
                id="Absolute player path (POSIX)",
            ),
            pytest.param(
                dict(path=Path("C:\\path\\to\\player.exe")),
                ["C:\\path\\to\\player.exe", "-"],
                marks=pytest.mark.windows_only,
                id="Absolute player path (Windows)",
            ),
        ],
        indirect=True,
    )
    def test_argv(self):
        pass

    # noinspection PyTestParametrized
    @pytest.mark.usefixtures("playerargs", "_playerargv")
    @pytest.mark.parametrize(
        ("playerargs", "_playerargv"),
        [
            pytest.param(
                dict(path=Path("player")),
                ["player", "-"],
                id="stdin",
            ),
            pytest.param(
                dict(path=Path("player"), namedpipe=Mock(path="namedpipe")),
                ["player", "namedpipe"],
                id="namedpipe",
            ),
            pytest.param(
                dict(path=Path("player"), filename="https://localhost:65535/-._~:/?#[]@!$&'()*+,;="),
                ["player", "https://localhost:65535/-._~:/?#[]@!$&'()*+,;="],
                id="filename",
            ),
            pytest.param(
                dict(path=Path("player"), http=Mock(url="https://localhost:65535/")),
                ["player", "https://localhost:65535/"],
                id="http",
            ),
        ],
        indirect=True,
    )
    def test_input(self):
        pass

    @pytest.mark.parametrize(
        ("playerpath", "playerargs", "playerargsclass"),
        [
            pytest.param(
                "player",
                "",
                PlayerArgs,
                id="Unknown player",
            ),
            pytest.param(
                "/usr/bin/flatpak",
                "upgrade",
                PlayerArgs,
                marks=pytest.mark.posix_only,
                id="Invalid flatpak command",
            ),
            pytest.param(
                "/usr/bin/flatpak",
                "run unknown",
                PlayerArgs,
                marks=pytest.mark.posix_only,
                id="Unknown flatpak player",
            ),
            pytest.param(
                "vlc",
                "",
                PlayerArgsVLC,
                id="VLC",
            ),
            pytest.param(
                "vlc.exe",
                "",
                PlayerArgs,
                marks=pytest.mark.posix_only,
                id="VLC with .exe file extension (POSIX)",
            ),
            pytest.param(
                "vlc.exe",
                "",
                PlayerArgsVLC,
                marks=pytest.mark.windows_only,
                id="VLC with .exe file extension (Windows)",
            ),
            pytest.param(
                "/usr/bin/vlc",
                "",
                PlayerArgsVLC,
                marks=pytest.mark.posix_only,
                id="VLC with absolute path (POSIX)",
            ),
            pytest.param(
                "/usr/bin/flatpak",
                '--system run --cwd="./foo bar" org.videolan.VLC --play-and-exit',
                PlayerArgsVLC,
                marks=pytest.mark.posix_only,
                id="VLC flatpak",
            ),
            pytest.param(
                "C:\\Program Files\\VideoLAN\\VLC\\vlc",
                "",
                PlayerArgsVLC,
                marks=pytest.mark.windows_only,
                id="VLC with absolute path (Windows)",
            ),
            pytest.param(
                "C:\\Program Files\\VideoLAN\\VLC\\vlc.exe",
                "",
                PlayerArgsVLC,
                marks=pytest.mark.windows_only,
                id="VLC with absolute path and file extension (Windows)",
            ),
            # don't repeat path and file extension tests
            # MPV
            pytest.param(
                "mpv",
                "",
                PlayerArgsMPV,
                id="MPV",
            ),
            pytest.param(
                "/usr/bin/flatpak",
                '--system run --cwd="./foo bar" io.mpv.Mpv --window-maximized',
                PlayerArgsMPV,
                marks=pytest.mark.posix_only,
                id="mpv flatpak",
            ),
            # Potplayer
            pytest.param(
                "potplayer",
                "",
                PlayerArgsPotplayer,
                id="Potplayer (potplayer)",
            ),
            pytest.param(
                "potplayermini",
                "",
                PlayerArgsPotplayer,
                id="Potplayer (potplayermini)",
            ),
            pytest.param(
                "potplayermini64",
                "",
                PlayerArgsPotplayer,
                id="Potplayer (potplayermini64)",
            ),
        ],
    )
    def test_knownplayer(self, playerpath: str, playerargs: str, playerargsclass: type[PlayerArgs]):
        assert isinstance(PlayerArgs(path=Path(playerpath), args=playerargs), playerargsclass)

    # noinspection PyTestParametrized
    @pytest.mark.usefixtures("playerargs", "_playerargv")
    @pytest.mark.parametrize(
        ("playerargs", "_playerargv"),
        [
            pytest.param(
                dict(path=Path("vlc"), namedpipe=Mock(path="namedpipe")),
                ["vlc", "namedpipe"],
                marks=pytest.mark.posix_only,
                id="VLC named pipe (POSIX)",
            ),
            pytest.param(
                dict(path=Path("vlc"), namedpipe=Mock(path="namedpipe")),
                ["vlc", "stream://\\namedpipe"],
                marks=pytest.mark.windows_only,
                id="VLC named pipe (Windows)",
            ),
            pytest.param(
                dict(path=Path("mpv"), namedpipe=Mock(path="namedpipe")),
                ["mpv", "namedpipe"],
                marks=pytest.mark.posix_only,
                id="MPV named pipe (POSIX)",
            ),
            pytest.param(
                dict(path=Path("mpv"), namedpipe=Mock(path="namedpipe")),
                ["mpv", "file://namedpipe"],
                marks=pytest.mark.windows_only,
                id="MPV named pipe (Windows)",
            ),
        ],
        indirect=True,
    )
    def test_knownplayer_input(self):
        pass

    # noinspection PyTestParametrized
    @pytest.mark.usefixtures("playerargs", "_playerargv")
    @pytest.mark.parametrize(
        ("playerargs", "_playerargv"),
        [
            pytest.param(
                dict(path=Path("player"), title="foo bar"),
                ["player", "-"],
                id="No title on unknown player",
            ),
            pytest.param(
                dict(path=Path("vlc"), title="foo bar: $a - $t"),
                ["vlc", "--input-title-format", "foo bar: $$a - $$t", "-"],
                id="VLC title",
            ),
            pytest.param(
                dict(path=Path("mpv"), title="foo bar"),
                ["mpv", "--force-media-title=foo bar", "-"],
                id="MPV title",
            ),
            pytest.param(
                dict(path=Path("potplayer"), title="foo bar"),
                ["potplayer", "-"],
                id="Potplayer title (stdin - no title)",
            ),
            pytest.param(
                dict(path=Path("potplayer"), title="foo bar", namedpipe=Mock(path="namedpipe")),
                ["potplayer", "namedpipe\\foo bar"],
                id="Potplayer title (namedpipe)",
            ),
            pytest.param(
                dict(path=Path("vlc"), args="param1 {playertitleargs} param2", title="foo bar"),
                ["vlc", "param1", "--input-title-format", "foo bar", "param2", "-"],
                id="Explicit playertitleargs variable",
            ),
            pytest.param(
                dict(path=Path("vlc"), args="param1{playertitleargs}param2", title="foo bar"),
                ["vlc", "param1--input-title-format", "foo barparam2", "-"],
                id="Explicit playertitleargs variable with improper usage (correct tokenization)",
            ),
            pytest.param(
                dict(
                    path=Path("flatpak"),
                    args="run unknown",
                    title="foo bar",
                ),
                [
                    "flatpak",
                    "run",
                    "unknown",
                    "-",
                ],
                id="flatpak-unknown",
            ),
            pytest.param(
                dict(
                    path=Path("flatpak"),
                    args='--system run --cwd="./foo bar" io.mpv.Mpv --fullscreen',
                    title="foo bar",
                ),
                [
                    "flatpak",
                    "--system",
                    "run",
                    "--cwd=./foo bar",
                    "io.mpv.Mpv",
                    "--force-media-title=foo bar",
                    "--fullscreen",
                    "-",
                ],
                id="flatpak-mpv",
            ),
            pytest.param(
                dict(
                    path=Path("flatpak"),
                    args='--system run --cwd="./foo bar" io.mpv.Mpv --fullscreen {playertitleargs} --volume=100',
                    title="foo bar",
                ),
                [
                    "flatpak",
                    "--system",
                    "run",
                    "--cwd=./foo bar",
                    "io.mpv.Mpv",
                    "--fullscreen",
                    "--force-media-title=foo bar",
                    "--volume=100",
                    "-",
                ],
                id="flatpak-mpv-playertitleargs",
            ),
        ],
        indirect=True,
    )
    def test_title(self):
        pass


# TODO: refactor PlayerOutput and write proper tests
class TestPlayerOutput:
    @pytest.fixture(autouse=True)
    def _os_environ(self, os_environ: dict[str, str]):
        os_environ["FAKE"] = "ENVIRONMENT"
        yield
        assert sorted(os_environ.keys()) == ["FAKE"], "Doesn't pollute the os.environ dict with custom env vars"

    @pytest.fixture()
    def playeroutput(self, request: pytest.FixtureRequest):
        with patch("streamlink_cli.output.player.sleep"):
            playeroutput = PlayerOutput(**getattr(request, "param", {}))
            yield playeroutput
            playeroutput.close()

    @pytest.fixture()
    def mock_which(self, request: pytest.FixtureRequest, playeroutput: PlayerOutput):
        with patch("streamlink_cli.output.player.which", return_value=getattr(request, "param", None)) as mock_which:
            yield mock_which

    @pytest.fixture()
    def mock_popen(self, playeroutput: PlayerOutput):
        with (
            patch("streamlink_cli.output.player.sleep"),
            patch("subprocess.Popen", return_value=Mock(poll=Mock(side_effect=Mock(return_value=None)))) as mock_popen,
        ):
            yield mock_popen

    @pytest.mark.parametrize(
        ("playeroutput", "mock_which", "args", "env", "logmessage"),
        [
            pytest.param(
                {"path": Path("player"), "args": "param1 param2", "call": False},
                "/resolved/player",
                ["/resolved/player", "param1", "param2", "-"],
                {"FAKE": "ENVIRONMENT"},
                "Opening subprocess: ['/resolved/player', 'param1', 'param2', '-']",
                id="Without custom env vars",
            ),
            pytest.param(
                {"path": Path("player"), "args": "param1 param2", "env": [("VAR1", "abc"), ("VAR2", "def")], "call": False},
                "/resolved/player",
                ["/resolved/player", "param1", "param2", "-"],
                {"FAKE": "ENVIRONMENT", "VAR1": "abc", "VAR2": "def"},
                "Opening subprocess: ['/resolved/player', 'param1', 'param2', '-'], env: {'VAR1': 'abc', 'VAR2': 'def'}",
                id="With custom env vars",
            ),
        ],
        indirect=["playeroutput", "mock_which"],
    )
    def test_open_popen_parameters(
        self,
        caplog: pytest.LogCaptureFixture,
        playeroutput: PlayerOutput,
        mock_which: Mock,
        mock_popen: Mock,
        args: Sequence[str],
        env: Mapping[str, str],
        logmessage: str,
    ):
        caplog.set_level(1, "streamlink")

        assert getattr(playeroutput, "player", None) is None

        playeroutput.open()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.cli.output", "debug", logmessage),
        ]
        assert mock_which.call_args_list == [call("player")]
        assert mock_popen.call_args_list == [
            call(
                args,
                bufsize=0,
                env=env,
                stdin=subprocess.PIPE,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            ),
        ]
        assert playeroutput.player is not None
        assert not playeroutput.player.terminate.called  # type: ignore

        playeroutput.close()
        assert playeroutput.player.terminate.call_count == 1  # type: ignore

    @pytest.mark.parametrize(
        ("playeroutput", "mock_which", "expected", "warns"),
        [
            pytest.param(
                dict(path=Path("foo")),
                "foo",
                nullcontext(),
                False,
                id="Player found",
            ),
            pytest.param(
                dict(path=Path("foo")),
                None,
                pytest.raises(FileNotFoundError, match=r"^Player executable not found$"),
                False,
                id="Player not found",
            ),
            pytest.param(
                dict(path=Path('"foo bar"')),
                None,
                pytest.raises(FileNotFoundError, match=r"^Player executable not found$"),
                True,
                id="Player not found with quotation warning",
            ),
        ],
        indirect=["playeroutput", "mock_which"],
    )
    def test_open_error(
        self,
        recwarn: pytest.WarningsRecorder,
        playeroutput: PlayerOutput,
        mock_which: Mock,
        mock_popen: Mock,
        expected: AbstractContextManager,
        warns: bool,
    ):
        with expected:
            playeroutput.open()
        assert any(record.category is StreamlinkWarning for record in recwarn.list) is warns



from streamlink.plugins.vkvideolive import VKvideolive
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlVKvideolive(PluginCanHandleUrl):
    __plugin__ = VKvideolive

    should_match = [
        # live
        "https://live.vkvideo.ru/CHANNEL",
        "https://live.vkplay.ru/CHANNEL",
        "https://vkplay.live/CHANNEL",
        # VOD
        "https://live.vkvideo.ru/CHANNEL/record/VOD",
        "https://live.vkplay.ru/CHANNEL/record/VOD",
        "https://vkplay.live/CHANNEL/record/VOD",
    ]

    should_not_match = [
        "https://live.vkvideo.ru/",
        "https://live.vkplay.ru/",
        "https://vkplay.live/",
        "https://support.vkplay.ru/vkp_live",
        "https://vkplay.live/app/settings/edit",
        "https://vkplay.live/app/settings/external-apps",
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import streamlink_cli.main


if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    "argv",
    [
        pytest.param(
            ["--http-cookies-file", "doesnotexist"],
            id="http-cookies-file",
        ),
    ],
    indirect=["argv"],
)
def test_failure(monkeypatch: pytest.MonkeyPatch, tmp_path: Path, capsys: pytest.CaptureFixture[str], argv: list[str]):
    monkeypatch.chdir(tmp_path)
    with pytest.raises(SystemExit) as exc_info:
        streamlink_cli.main.main()
    assert exc_info.value.code == 1
    stdout, stderr = capsys.readouterr()
    assert stdout == ""
    assert stderr == f"Error while loading cookies from file: '{tmp_path / 'doesnotexist'}' is not a valid cookies file path\n"



import sys
from textwrap import dedent


CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 10)

# This check and everything above must remain compatible with older Python versions
if CURRENT_PYTHON < REQUIRED_PYTHON:
    raise SystemExit(
        dedent(
            """
                ========================================================
                               Unsupported Python version
                ========================================================
                Streamlink requires at least Python {}.{},
                but you're trying to install it on Python {}.{}.

                This may be because you are using a version of pip that
                doesn't understand the python_requires classifier.
                Make sure you have pip >= 9.0 and setuptools >= 24.2
            """,
        )
        .strip(" \n")
        .format(*REQUIRED_PYTHON, *CURRENT_PYTHON),
    )


from pathlib import Path  # noqa: E402
from typing import TYPE_CHECKING  # noqa: E402


if TYPE_CHECKING:
    from collections.abc import Sequence


def is_wheel_for_windows(argv):
    if "bdist_wheel" in argv:
        names = ["win32", "win-amd64", "cygwin"]
        length = len(argv)
        for pos in range(argv.index("bdist_wheel") + 1, length):
            if argv[pos] == "--plat-name" and pos + 1 < length:
                return argv[pos + 1] in names
            elif argv[pos][:12] == "--plat-name=":
                return argv[pos][12:] in names
    return False


entry_points = {
    "console_scripts": ["streamlink=streamlink_cli.main:main"],
}

if is_wheel_for_windows(sys.argv):
    entry_points["gui_scripts"] = ["streamlinkw=streamlink_cli.main:main"]


# optional data files
data_files: "list[tuple[str, Sequence[str]]]" = [  # noqa: UP037
    # shell completions:
    #  requires pre-built completion files via shtab ("build" dependency group)
    #  `./script/build-shell-completions.sh`
    ("share/bash-completion/completions", ["completions/bash/streamlink"]),
    ("share/zsh/site-functions", ["completions/zsh/_streamlink"]),
    # man page:
    #  requires the pre-built man page file via sphinx ("docs" dependency group)
    #  `make --directory=docs clean man`
    ("share/man/man1", ["docs/_build/man/streamlink.1"]),
]
data_files = [
    (destdir, [file for file in srcfiles if Path(file).exists()])
    for destdir, srcfiles in data_files
]  # fmt: skip


if __name__ == "__main__":
    sys.path.insert(0, str(Path(__file__).parent))

    from build_backend.commands import cmdclass
    from setuptools import Command, setup  # noqa: TC002

    try:
        # versioningit is only required when building from git (see pyproject.toml)
        from versioningit import get_cmdclasses
    except ImportError:  # pragma: no cover

        def get_cmdclasses(bases: "dict[str, type[Command]] | None" = None) -> "dict[str, type[Command]]":  # noqa: UP037
            return bases or {}

    setup(
        cmdclass=get_cmdclasses(cmdclass),
        entry_points=entry_points,
        data_files=data_files,
        version="8.4.0",
    )



from streamlink.plugins.euronews import Euronews
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlEuronews(PluginCanHandleUrl):
    __plugin__ = Euronews

    should_match_groups = [
        (("live", "https://euronews.com/live"), {}),
        (("live", "https://www.euronews.com/live"), {"subdomain": "www"}),
        (("live", "https://fr.euronews.com/live"), {"subdomain": "fr"}),
        (("live", "https://de.euronews.com/live"), {"subdomain": "de"}),
        (("live", "https://it.euronews.com/live"), {"subdomain": "it"}),
        (("live", "https://es.euronews.com/live"), {"subdomain": "es"}),
        (("live", "https://pt.euronews.com/live"), {"subdomain": "pt"}),
        (("live", "https://ru.euronews.com/live"), {"subdomain": "ru"}),
        (("live", "https://tr.euronews.com/live"), {"subdomain": "tr"}),
        (("live", "https://gr.euronews.com/live"), {"subdomain": "gr"}),
        (("live", "https://hu.euronews.com/live"), {"subdomain": "hu"}),
        (("live", "https://fa.euronews.com/live"), {"subdomain": "fa"}),
        (("live", "https://arabic.euronews.com/live"), {"subdomain": "arabic"}),
        (
            ("vod", "https://www.euronews.com/video/2025/02/28/latest-news-bulletin-february-28th-midday"),
            {},
        ),
        (
            ("vod", "https://www.euronews.com/my-europe/2025/02/25/whats-at-stake-for-europe-after-the-german-election"),
            {},
        ),
        (
            ("vod", "https://de.euronews.com/my-europe/2025/02/24/frankreichs-prasident-macron-wird-trump-treffen"),
            {},
        ),
    ]

    should_not_match = [
        "https://euronews.com/",
        "https://www.euronews.com/",
    ]



from __future__ import annotations

from contextlib import nullcontext
from datetime import datetime, timezone
from typing import TYPE_CHECKING
from unittest.mock import ANY, Mock, call

import freezegun
import pytest
from lxml.etree import ParseError

from streamlink.exceptions import PluginError
from streamlink.stream.dash import MPD, DASHSegment, DASHStream, DASHStreamWorker, MPDParsingError
from streamlink.stream.dash.dash import log
from streamlink.utils.parse import parse_xml as original_parse_xml
from tests.resources import text, xml


if TYPE_CHECKING:
    from collections.abc import Generator, Iterator

    import requests_mock as rm

    from streamlink.session import Streamlink
    from streamlink.stream.dash import DASHStreamReader


does_not_raise = nullcontext()


@pytest.fixture()
def timestamp():
    with freezegun.freeze_time("2000-01-01T00:00:00Z"):
        yield datetime.now(timezone.utc)


def test_logger_name():
    assert log.name == "streamlink.stream.dash"


class TestDASHStreamParseManifest:
    @pytest.fixture(autouse=True)
    def _response(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker):
        response = requests_mock.register_uri("GET", "http://test/manifest.mpd", **getattr(request, "param", {}))
        called_once = "nomockedhttprequest" not in request.keywords
        yield
        assert (response.call_count == 1) is called_once

    @pytest.fixture()
    def parse_xml(self, monkeypatch: pytest.MonkeyPatch):
        parse_xml = Mock(return_value=Mock())
        monkeypatch.setattr("streamlink.stream.dash.dash.parse_xml", parse_xml)
        return parse_xml

    @pytest.fixture()
    def mpd(self, monkeypatch: pytest.MonkeyPatch, parse_xml: Mock):
        mpd = Mock()
        monkeypatch.setattr("streamlink.stream.dash.dash.MPD", mpd)
        return mpd

    @pytest.mark.parametrize(
        ("se_parse_xml", "se_mpd"),
        [
            (ParseError, None),
            (None, MPDParsingError),
        ],
    )
    def test_parse_fail(self, session: Streamlink, mpd: Mock, parse_xml: Mock, se_parse_xml, se_mpd):
        parse_xml.side_effect = se_parse_xml
        mpd.side_effect = se_mpd

        with pytest.raises(PluginError) as cm:
            DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert str(cm.value).startswith("Failed to parse MPD manifest: ")

    def test_video_only(self, session: Streamlink, mpd: Mock):
        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="video/mp4", height=720),
                Mock(id="2", contentProtections=None, mimeType="video/mp4", height=1080),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert sorted(streams.keys()) == sorted(["720p", "1080p"])

    def test_audio_only(self, session: Streamlink, mpd: Mock):
        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="audio/mp4", bandwidth=128.0, lang="en"),
                Mock(id="2", contentProtections=None, mimeType="audio/mp4", bandwidth=256.0, lang="en"),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert sorted(streams.keys()) == sorted(["a128k", "a256k"])

    @pytest.mark.parametrize(
        ("with_video_only", "with_audio_only", "expected"),
        [
            pytest.param(
                False,
                False,
                ["720p+a128k", "720p+a256k", "1080p+a128k", "1080p+a256k"],
                id="Only muxed streams",
            ),
            pytest.param(
                True,
                False,
                ["720p", "720p+a128k", "720p+a256k", "1080p", "1080p+a128k", "1080p+a256k"],
                id="With video-only streams",
            ),
            pytest.param(
                False,
                True,
                ["a128k", "a256k", "720p+a128k", "720p+a256k", "1080p+a128k", "1080p+a256k"],
                id="With audio-only streams",
            ),
            pytest.param(
                True,
                True,
                ["a128k", "a256k", "720p", "720p+a128k", "720p+a256k", "1080p", "1080p+a128k", "1080p+a256k"],
                id="With video-only and audio-only streams",
            ),
        ],
    )
    def test_with_videoaudio_only(
        self,
        session: Streamlink,
        mpd: Mock,
        with_video_only: bool,
        with_audio_only: bool,
        expected: list[str],
    ):
        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="video/mp4", height=720),
                Mock(id="2", contentProtections=None, mimeType="video/mp4", height=1080),
                Mock(id="3", contentProtections=None, mimeType="audio/mp4", bandwidth=128.0, lang="en"),
                Mock(id="4", contentProtections=None, mimeType="audio/mp4", bandwidth=256.0, lang="en"),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(
            session,
            "http://test/manifest.mpd",
            with_video_only=with_video_only,
            with_audio_only=with_audio_only,
        )
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert list(streams.keys()) == expected

    def test_audio_single(self, session: Streamlink, mpd: Mock):
        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="video/mp4", height=720),
                Mock(id="2", contentProtections=None, mimeType="video/mp4", height=1080),
                Mock(id="3", contentProtections=None, mimeType="audio/aac", bandwidth=128.0, lang="en"),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert sorted(streams.keys()) == sorted(["720p", "1080p"])

    def test_audio_multi(self, session: Streamlink, mpd: Mock):
        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="video/mp4", height=720),
                Mock(id="2", contentProtections=None, mimeType="video/mp4", height=1080),
                Mock(id="3", contentProtections=None, mimeType="audio/aac", bandwidth=128.0, lang="en"),
                Mock(id="4", contentProtections=None, mimeType="audio/aac", bandwidth=256.0, lang="en"),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert sorted(streams.keys()) == sorted(["720p+a128k", "1080p+a128k", "720p+a256k", "1080p+a256k"])

    def test_audio_multi_lang(self, session: Streamlink, mpd: Mock):
        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="video/mp4", height=720),
                Mock(id="2", contentProtections=None, mimeType="video/mp4", height=1080),
                Mock(id="3", contentProtections=None, mimeType="audio/aac", bandwidth=128.0, lang="en"),
                Mock(id="4", contentProtections=None, mimeType="audio/aac", bandwidth=128.0, lang="es"),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert sorted(streams.keys()) == sorted(["720p", "1080p"])
        assert getattr(streams["720p"].audio_representation, "lang", None) == "en"
        assert getattr(streams["1080p"].audio_representation, "lang", None) == "en"

    def test_audio_multi_lang_alpha3(self, session: Streamlink, mpd: Mock):
        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="video/mp4", height=720),
                Mock(id="2", contentProtections=None, mimeType="video/mp4", height=1080),
                Mock(id="3", contentProtections=None, mimeType="audio/aac", bandwidth=128.0, lang="eng"),
                Mock(id="4", contentProtections=None, mimeType="audio/aac", bandwidth=128.0, lang="spa"),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert sorted(streams.keys()) == sorted(["720p", "1080p"])
        assert getattr(streams["720p"].audio_representation, "lang", None) == "eng"
        assert getattr(streams["1080p"].audio_representation, "lang", None) == "eng"

    def test_audio_invalid_lang(self, session: Streamlink, mpd: Mock):
        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="video/mp4", height=720),
                Mock(id="2", contentProtections=None, mimeType="video/mp4", height=1080),
                Mock(id="3", contentProtections=None, mimeType="audio/aac", bandwidth=128.0, lang="en_no_voice"),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert sorted(streams.keys()) == sorted(["720p", "1080p"])
        assert getattr(streams["720p"].audio_representation, "lang", None) == "en_no_voice"
        assert getattr(streams["1080p"].audio_representation, "lang", None) == "en_no_voice"

    def test_audio_multi_lang_locale(self, monkeypatch: pytest.MonkeyPatch, session: Streamlink, mpd: Mock):
        session.set_option("locale", "es_ES")

        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="video/mp4", height=720),
                Mock(id="2", contentProtections=None, mimeType="video/mp4", height=1080),
                Mock(id="3", contentProtections=None, mimeType="audio/aac", bandwidth=128.0, lang="en"),
                Mock(id="4", contentProtections=None, mimeType="audio/aac", bandwidth=128.0, lang="es"),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert sorted(streams.keys()) == sorted(["720p", "1080p"])
        assert getattr(streams["720p"].audio_representation, "lang", None) == "es"
        assert getattr(streams["1080p"].audio_representation, "lang", None) == "es"

    # Verify the fix for https://github.com/streamlink/streamlink/issues/3365
    def test_duplicated_resolutions(self, session: Streamlink, mpd: Mock):
        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="video/mp4", height=1080, bandwidth=128.0),
                Mock(id="2", contentProtections=None, mimeType="video/mp4", height=1080, bandwidth=64.0),
                Mock(id="3", contentProtections=None, mimeType="video/mp4", height=1080, bandwidth=32.0),
                Mock(id="4", contentProtections=None, mimeType="video/mp4", height=720),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert sorted(streams.keys()) == sorted(["720p", "1080p", "1080p_alt", "1080p_alt2"])

    # Verify the fix for https://github.com/streamlink/streamlink/issues/4217
    def test_duplicated_resolutions_sorted_bandwidth(self, session: Streamlink, mpd: Mock):
        adaptationset = Mock(
            contentProtections=None,
            representations=[
                Mock(id="1", contentProtections=None, mimeType="video/mp4", height=1080, bandwidth=64.0),
                Mock(id="2", contentProtections=None, mimeType="video/mp4", height=1080, bandwidth=128.0),
                Mock(id="3", contentProtections=None, mimeType="video/mp4", height=1080, bandwidth=32.0),
            ],
        )
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert getattr(streams["1080p"].video_representation, "bandwidth", None) == pytest.approx(128.0)
        assert getattr(streams["1080p_alt"].video_representation, "bandwidth", None) == pytest.approx(64.0)
        assert getattr(streams["1080p_alt2"].video_representation, "bandwidth", None) == pytest.approx(32.0)

    @pytest.mark.parametrize(
        "adaptationset",
        [
            pytest.param(
                Mock(contentProtections="DRM", representations=[]),
                id="adaptationset",
            ),
            pytest.param(
                Mock(contentProtections=None, representations=[Mock(id="1", contentProtections="DRM", height=1080)]),
                id="representation",
            ),
        ],
    )
    @pytest.mark.parametrize(
        ("session", "raises", "logrecords"),
        [
            pytest.param(
                {},
                pytest.raises(PluginError, match=r" is protected by DRM$"),
                [],
                id="no-passthrough-encrypted",
            ),
            pytest.param(
                {"stream-passthrough-encrypted": True},
                does_not_raise,
                [
                    (
                        "streamlink.stream.dash",
                        "warning",
                        "http://test/manifest.mpd is protected by DRM and won't be decrypted",
                    ),
                ],
                id="passthrough-encrypted",
            ),
        ],
        indirect=["session"],
    )
    def test_contentprotection(
        self,
        caplog: pytest.LogCaptureFixture,
        session: Streamlink,
        raises: nullcontext,
        adaptationset: Mock,
        mpd: Mock,
        logrecords: list,
    ):
        mpd.return_value = Mock(periods=[Mock(adaptationSets=[adaptationset])])

        with raises:
            DASHStream.parse_manifest(session, "http://test/manifest.mpd")

        assert [(record.name, record.levelname, record.message) for record in caplog.records] == logrecords

    @pytest.mark.nomockedhttprequest()
    def test_string(self, session: Streamlink, mpd: Mock, parse_xml: Mock):
        with text("dash/test_9.mpd") as mpd_txt:
            test_manifest = mpd_txt.read()
        parse_xml.side_effect = original_parse_xml
        mpd.side_effect = MPD

        streams = DASHStream.parse_manifest(session, test_manifest)
        assert mpd.call_args_list == [call(ANY)]
        assert list(streams.keys()) == ["480p"]

    # TODO: Move this test to test_dash_parser and properly test segment URLs.
    #       This test currently achieves nothing... (manifest fixture added in 7aada92)
    def test_segments_number_time(self, session: Streamlink, mpd: Mock):
        with xml("dash/test_9.mpd") as mpd_xml:
            mpd.return_value = MPD(mpd_xml, base_url="http://test/manifest.mpd", url="http://test/manifest.mpd")

        streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd")
        assert mpd.call_args_list == [call(ANY, url="http://test/manifest.mpd", base_url="http://test/manifest.mpd")]
        assert list(streams.keys()) == ["480p"]

    @pytest.mark.parametrize(
        ("period", "raises"),
        [
            pytest.param(0, does_not_raise, id="index-0"),
            pytest.param(1, does_not_raise, id="index-1"),
            pytest.param(2, does_not_raise, id="index-2"),
            pytest.param("p1", does_not_raise, id="id-p1"),
            pytest.param("p2", does_not_raise, id="id-p2"),
            pytest.param(3, pytest.raises(PluginError, match=r"^DASH period 3 not found\."), id="error-index"),
            pytest.param("p3", pytest.raises(PluginError, match=r"^DASH period 'p3' not found\."), id="error-id"),
        ],
    )
    def test_period_selection(
        self,
        caplog: pytest.LogCaptureFixture,
        session: Streamlink,
        mpd: Mock,
        period: int | str,
        raises: nullcontext,
    ):
        caplog.set_level(1, "streamlink")

        with xml("dash/test_period_selection.mpd") as mpd_xml:
            mpd.return_value = MPD(mpd_xml, base_url="http://test/manifest.mpd", url="http://test/manifest.mpd")

        with raises:
            streams = DASHStream.parse_manifest(session, "http://test/manifest.mpd", period=period)
            assert streams

        records = [(record.name, record.levelname, record.message) for record in caplog.records]
        assert ("streamlink.stream.dash", "debug", "Available DASH periods: 0, 1 (id='p1'), 2 (id='p2')") in records


class TestDASHStreamOpen:
    @pytest.fixture()
    def reader(self, monkeypatch: pytest.MonkeyPatch):
        reader = Mock()
        monkeypatch.setattr("streamlink.stream.dash.dash.DASHStreamReader", reader)
        return reader

    @pytest.fixture()
    def muxer(self, monkeypatch: pytest.MonkeyPatch):
        muxer = Mock()
        muxer.is_usable.return_value = True
        monkeypatch.setattr("streamlink.stream.dash.dash.FFMPEGMuxer", muxer)
        return muxer

    def test_stream_open_video_only(self, session: Streamlink, timestamp: datetime, muxer: Mock, reader: Mock):
        rep_video = Mock(ident=(None, None, "1"), mimeType="video/mp4")
        stream = DASHStream(session, Mock(), rep_video)
        stream.open()

        assert reader.call_args_list == [call(stream, rep_video, timestamp, name="video")]
        assert reader().open.call_count == 1
        assert muxer.call_args_list == []

    def test_stream_open_video_audio(self, session: Streamlink, timestamp: datetime, muxer: Mock, reader: Mock):
        rep_video = Mock(ident=(None, None, "1"), mimeType="video/mp4")
        rep_audio = Mock(ident=(None, None, "2"), mimeType="audio/mp3", lang="en")

        mock_reader_video = Mock()
        mock_reader_audio = Mock()
        readers = {rep_video: mock_reader_video, rep_audio: mock_reader_audio}
        reader.side_effect = lambda _stream, _representation, _timestamp, *_, **__: readers[_representation]

        stream = DASHStream(session, Mock(), rep_video, rep_audio)
        stream.open()

        assert reader.call_args_list == [
            call(stream, rep_video, timestamp, name="video"),
            call(stream, rep_audio, timestamp, name="audio"),
        ]
        assert mock_reader_video.open.call_count == 1
        assert mock_reader_audio.open.call_count == 1
        assert muxer.call_args_list == [call(session, mock_reader_video, mock_reader_audio, copyts=True)]


class TestDASHStreamWorker:
    @pytest.fixture()
    def mock_time(self, monkeypatch: pytest.MonkeyPatch) -> Mock:
        mock = Mock(return_value=1)
        monkeypatch.setattr("streamlink.stream.dash.dash.time", mock)
        return mock

    @pytest.fixture(autouse=True)
    def mock_wait(self, monkeypatch: pytest.MonkeyPatch) -> Mock:
        mock = Mock(return_value=True)
        monkeypatch.setattr("streamlink.stream.dash.dash.DASHStreamWorker.wait", mock)
        return mock

    @pytest.fixture()
    def segments(self) -> list[DASHSegment]:
        return [
            DASHSegment(uri="init_segment", num=-1, duration=0.0),
            DASHSegment(uri="first_segment", num=0, duration=2.0),
            DASHSegment(uri="second_segment", num=1, duration=3.0),
            DASHSegment(uri="third_segment", num=2, duration=5.0),
            DASHSegment(uri="fourth_segment", num=3, duration=8.0),
            DASHSegment(uri="fifth_segment", num=4, duration=13.0),
        ]

    @pytest.fixture()
    def mpd(self) -> Mock:
        representation = Mock(
            ident=(None, None, "1"),
            mimeType="video/mp4",
            height=720,
        )
        adaptationset = Mock(
            contentProtections=None,
            representations=[representation],
        )
        period = Mock(
            duration=Mock(total_seconds=Mock(return_value=0)),
            adaptationSets=[adaptationset],
        )
        representation.period = period

        return Mock(
            publishTime=1,
            minimumUpdatePeriod=Mock(total_seconds=Mock(return_value=0)),
            periods=[period],
            get_representation=Mock(return_value=representation),
        )

    @pytest.fixture()
    def representation(self, mpd) -> Mock:
        return mpd.periods[0].adaptationSets[0].representations[0]

    @pytest.fixture()
    def stream(self, request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch, session: Streamlink, mpd: Mock):
        options = getattr(request, "param", {})

        monkeypatch.setattr(session.http, "request", Mock())
        monkeypatch.setattr(session.http, "xml", Mock())

        return DASHStream(session, mpd, **options)

    @pytest.fixture()
    def reader(self, session: Streamlink, stream: DASHStream, timestamp: datetime):
        return Mock(
            session=session,
            stream=stream,
            ident=(None, None, "1"),
            timestamp=timestamp,
        )

    @pytest.fixture()
    def worker(self, reader: DASHStreamReader):
        return DASHStreamWorker(reader)

    @staticmethod
    def _iter_segments(iter_segments: Generator[DASHSegment, bool, None]) -> Iterator[DASHSegment]:
        queued: bool | None = None
        try:
            while True:
                if queued is None:
                    yield next(iter_segments)
                else:
                    yield iter_segments.send(queued)
                queued = True
        except StopIteration:
            pass

    @staticmethod
    def _next_segments(worker: DASHStreamWorker, segment_iter: Iterator[DASHSegment], num: int) -> list[DASHSegment]:
        items = []
        for _ in range(num):
            segment = next(segment_iter)
            # fake worker.run() implementation
            worker.check_sequence_gap(segment)
            worker.sequence = segment.num + 1
            items.append(segment)

        return items

    def test_dynamic_reload(
        self,
        monkeypatch: pytest.MonkeyPatch,
        caplog: pytest.LogCaptureFixture,
        timestamp: datetime,
        worker: DASHStreamWorker,
        representation: Mock,
        segments: list[DASHSegment],
        mpd: Mock,
    ):
        caplog.set_level("INFO", "streamlink")

        mpd.dynamic = True
        mpd.type = "dynamic"
        monkeypatch.setattr("streamlink.stream.dash.dash.MPD", lambda *args, **kwargs: mpd)

        segment_iter = self._iter_segments(worker.iter_segments())

        representation.segments.return_value = segments[:2]
        assert self._next_segments(worker, segment_iter, 2) == segments[:2]
        assert representation.segments.call_args_list == [call(sequence=-1, init=True, timestamp=timestamp)]
        assert not worker._wait.is_set()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == []

        representation.segments.reset_mock()
        representation.segments.return_value = segments[3:]
        assert self._next_segments(worker, segment_iter, 3) == segments[3:]
        assert representation.segments.call_args_list == [call(), call(sequence=1, init=False, timestamp=None)]
        assert not worker._wait.is_set()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.stream.segmented",
                "warning",
                "Sequence gap of 1 segment at position 1. This is unsupported and will result in incoherent output data.",
            ),
        ]

    def test_dynamic_reload_missing_representation(
        self,
        monkeypatch: pytest.MonkeyPatch,
        caplog: pytest.LogCaptureFixture,
        timestamp: datetime,
        worker: DASHStreamWorker,
        representation: Mock,
        segments: list[DASHSegment],
        mpd: Mock,
    ):
        caplog.set_level("INFO", "streamlink")

        mpd.dynamic = True
        mpd.type = "dynamic"
        monkeypatch.setattr("streamlink.stream.dash.dash.MPD", lambda *args, **kwargs: mpd)

        segment_iter = self._iter_segments(worker.iter_segments())

        representation.segments.return_value = segments[:2]
        assert self._next_segments(worker, segment_iter, 2) == segments[:2]
        assert representation.segments.call_args_list == [call(sequence=-1, init=True, timestamp=timestamp)]
        assert not worker.closed
        assert not worker._wait.is_set()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == []

        mpd.get_representation.return_value = None
        with pytest.raises(StopIteration):
            self._next_segments(worker, segment_iter, 1)
        assert worker.closed
        assert worker._wait.is_set()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            (
                "streamlink.stream.dash",
                "error",
                "Failed to find matching DASH representation: (None, None, '1')",
            ),
        ]

    def test_static(
        self,
        worker: DASHStreamWorker,
        timestamp: datetime,
        representation: Mock,
        segments: list[DASHSegment],
        mpd: Mock,
    ):
        mpd.dynamic = False
        mpd.type = "static"

        representation.segments.return_value = segments
        assert list(self._iter_segments(worker.iter_segments())) == segments
        assert representation.segments.call_args_list == [call(sequence=-1, init=True, timestamp=timestamp)]
        assert worker._wait.is_set()

    # Verify the fix for https://github.com/streamlink/streamlink/issues/2873
    @pytest.mark.parametrize(
        "period_duration",
        [
            0,
            204.32,
        ],
    )
    def test_static_refresh_wait(
        self,
        timestamp: datetime,
        mock_wait: Mock,
        mock_time: Mock,
        worker: DASHStreamWorker,
        representation: Mock,
        segments: list[DASHSegment],
        mpd: Mock,
        period_duration: float,
    ):
        mpd.dynamic = False
        mpd.type = "static"
        mpd.periods[0].duration.total_seconds.return_value = period_duration

        representation.segments.return_value = segments
        assert list(self._iter_segments(worker.iter_segments())) == segments
        assert representation.segments.call_args_list == [call(sequence=-1, init=True, timestamp=timestamp)]
        assert mock_wait.call_args_list == [call(5)]
        assert worker._wait.is_set()

    @pytest.mark.parametrize(
        ("stream", "session"),
        [
            pytest.param(
                {"duration": 5.0},
                {},
                id="duration keyword",
            ),
            pytest.param(
                {},
                {"stream-segmented-duration": 5.0},
                id="stream-segmented-duration session option",
            ),
            pytest.param(
                {"duration": 5.0},
                {"stream-segmented-duration": 2.0},
                id="duration keyword priority",
            ),
        ],
        indirect=["stream", "session"],
    )
    def test_duration(
        self,
        caplog: pytest.LogCaptureFixture,
        reader: Mock,
        worker: DASHStreamWorker,
        timestamp: datetime,
        representation: Mock,
        segments: list[DASHSegment],
        mpd: Mock,
        stream: DASHStream,
        session: Streamlink,
    ):
        caplog.set_level("INFO", "streamlink")

        mpd.dynamic = False
        mpd.type = "static"

        representation.segments.return_value = segments
        worker.run()

        assert [call_arg.args[0] for call_arg in reader.writer.put.call_args_list] == [*segments[0:3], None]
        assert representation.segments.call_args_list == [call(sequence=-1, init=True, timestamp=timestamp)]
        assert worker._wait.is_set()
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == [
            ("streamlink.stream.segmented", "info", "Stopping stream early after 5.00s"),
        ]



from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

import streamlink_cli.main
import tests


if TYPE_CHECKING:
    from streamlink.session import Streamlink


@pytest.fixture(autouse=True)
def session(session: Streamlink):
    session.plugins.load_path(Path(tests.__path__[0]) / "plugin")

    return session


@pytest.mark.parametrize(
    "argv",
    [pytest.param(["--player-passthrough=hls", "http://test.se", "hls"], id="no-player")],
    indirect=["argv"],
)
def test_no_player(monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str], argv: list):
    monkeypatch.setattr("streamlink_cli.argparser.find_default_player", lambda *_, **__: None)

    with pytest.raises(SystemExit) as exc_info:
        streamlink_cli.main.main()
    assert exc_info.value.code == 1

    out, _err = capsys.readouterr()
    assert re.search(r"error: The default player \(\w+\) does not seem to be installed\.", out)



from streamlink.plugins.nowtvtr import NowTVTR
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlFoxTR(PluginCanHandleUrl):
    __plugin__ = NowTVTR

    should_match = [
        "https://www.nowtv.com.tr/canli-yayin",
        "https://www.nowtv.com.tr/now-haber",
        "https://www.nowtv.com.tr/yayin-akisi",
        "https://www.nowtv.com.tr/Gaddar/bolumler",
        "https://www.nowtv.com.tr/Memet-Ozer-ile-Mutfakta/bolumler",
    ]



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import ANY, MagicMock

from streamlink.plugins.ustreamtv import UStreamTV
from tests.plugins import PluginCanHandleUrl


if TYPE_CHECKING:
    from streamlink import Streamlink


class TestPluginCanHandleUrlUStreamTV(PluginCanHandleUrl):
    __plugin__ = UStreamTV

    should_match_groups = [
        (
            "https://www.ustream.tv/nasahdtv",
            {},
        ),
        (
            "https://www.ustream.tv/channel/6540154",
            {"channel_id": "6540154"},
        ),
        (
            "https://www.ustream.tv/channel/id/6540154",
            {"channel_id": "6540154"},
        ),
        (
            "https://www.ustream.tv/embed/6540154",
            {"channel_id": "6540154"},
        ),
        (
            "https://www.ustream.tv/recorded/132041157",
            {"video_id": "132041157"},
        ),
        (
            "https://www.ustream.tv/embed/recorded/132041157",
            {"video_id": "132041157"},
        ),
        (
            "https://www.ustream.tv/combined-embed/6540154",
            {"combined_channel_id": "6540154"},
        ),
        (
            "https://www.ustream.tv/combined-embed/6540154/video/132041157",
            {"combined_channel_id": "6540154", "combined_video_id": "132041157"},
        ),
        (
            "https://video.ibm.com/nasahdtv",
            {},
        ),
        (
            "https://video.ibm.com/recorded/132041157",
            {"video_id": "132041157"},
        ),
    ]


class TestPluginUStreamTV:
    def test_arguments(self, session: Streamlink):
        from streamlink_cli.main import setup_plugin_args  # noqa: PLC0415

        parser = MagicMock()
        plugins = parser.add_argument_group("Plugin Options")
        group = parser.add_argument_group("UStreamTV", parent=plugins)

        session.plugins["ustreamtv"] = UStreamTV

        setup_plugin_args(session, parser)

        group.add_argument.assert_called_with("--ustream-password", metavar="PASSWORD", help=ANY)



from streamlink.plugins.tiktok import TikTok
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTikTok(PluginCanHandleUrl):
    __plugin__ = TikTok

    should_match_groups = [
        (("live", "https://www.tiktok.com/@LIVE"), {"channel": "LIVE"}),
        (("live", "https://www.tiktok.com/@LIVE/live"), {"channel": "LIVE"}),
        (("video", "https://www.tiktok.com/@VIDEO/video/0123456789"), {"channel": "VIDEO", "id": "0123456789"}),
    ]

    should_not_match = [
        "https://www.tiktok.com",
    ]



import importlib.util

import pytest

from streamlink.exceptions import StreamlinkDeprecationWarning


@pytest.mark.parametrize(
    "attr",
    [
        "DASHStream",
        "MuxedStream",
        "HLSStream",
        "MuxedHLSStream",
        "HTTPStream",
        "Stream",
        "StreamIO",
        "StreamIOIterWrapper",
        "StreamIOThreadWrapper",
        "StreamIOWrapper",
    ],
)
def test_deprecated(attr: str):
    spec = importlib.util.find_spec("streamlink.stream", "streamlink")
    assert spec
    assert spec.loader
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    with pytest.warns(StreamlinkDeprecationWarning):
        item = getattr(module, attr)

    assert item



from __future__ import annotations

from pathlib import Path
from string import ascii_lowercase as alphabet

import pytest

from streamlink_cli.utils.path import replace_chars, replace_path, truncate_path


@pytest.mark.parametrize("char", list(range(32)))
def test_replace_chars_unprintable(char: int):
    assert replace_chars(f"foo{chr(char)}{chr(char)}bar") == "foo_bar", "Replaces unprintable characters"


@pytest.mark.posix_only()
@pytest.mark.parametrize("char", "/".split())
def test_replace_chars_posix(char: str):
    assert replace_chars(f"foo{char}{char}bar") == "foo_bar", "Replaces multiple unsupported characters in a row"


@pytest.mark.windows_only()
@pytest.mark.parametrize("char", '\x7f"*/:<>?\\|'.split())
def test_replace_chars_windows(char: str):
    assert replace_chars(f"foo{char}{char}bar") == "foo_bar", "Replaces multiple unsupported characters in a row"


@pytest.mark.posix_only()
def test_replace_chars_posix_all():
    assert replace_chars("".join(chr(i) for i in range(32)) + "/") == "_"


@pytest.mark.windows_only()
def test_replace_chars_windows_all():
    assert replace_chars("".join(chr(i) for i in range(32)) + '\x7f"*/:<>?\\|') == "_"


@pytest.mark.posix_only()
def test_replace_chars_posix_override():
    all_chars = "".join(chr(i) for i in range(32)) + '\x7f"*:/<>?\\|'
    assert replace_chars(all_chars) == '_\x7f"*:_<>?\\|'
    assert replace_chars(all_chars, "posix") == '_\x7f"*:_<>?\\|'
    assert replace_chars(all_chars, "unix") == '_\x7f"*:_<>?\\|'
    assert replace_chars(all_chars, "windows") == "_"
    assert replace_chars(all_chars, "win32") == "_"


@pytest.mark.windows_only()
def test_replace_chars_windows_override():
    all_chars = "".join(chr(i) for i in range(32)) + '\x7f"*:/<>?\\|'
    assert replace_chars(all_chars) == "_"
    assert replace_chars(all_chars, "posix") == '_\x7f"*:_<>?\\|'
    assert replace_chars(all_chars, "unix") == '_\x7f"*:_<>?\\|'
    assert replace_chars(all_chars, "windows") == "_"
    assert replace_chars(all_chars, "win32") == "_"


def test_replace_chars_replacement():
    assert replace_chars("\x00", None, "+") == "+"


def test_replace_path():
    def mapper(s, *_):
        return dict(foo=".", bar="..").get(s, s)

    path = Path("foo", ".", "bar", "..", "baz")
    expected = Path("_", ".", "_", "..", "baz")
    assert replace_path(path, mapper) == expected, "Only replaces mapped parts which are in the special parts tuple"


@pytest.mark.posix_only()
@pytest.mark.parametrize("os_environ", [pytest.param({"HOME": "/home/foo"}, id="posix")], indirect=True)
def test_replace_path_expanduser_posix(os_environ):
    assert replace_path("~/bar", lambda s, *_: s) == Path("/home/foo/bar")
    assert replace_path("foo/bar", lambda s, *_: dict(foo="~").get(s, s)) == Path("~/bar")


@pytest.mark.windows_only()
@pytest.mark.parametrize("os_environ", [pytest.param({"USERPROFILE": "C:\\Users\\foo"}, id="windows")], indirect=True)
def test_replace_path_expanduser_windows(os_environ):
    assert replace_path("~\\bar", lambda s, *_: s) == Path("C:\\Users\\foo\\bar")
    assert replace_path("foo\\bar", lambda s, *_: dict(foo="~").get(s, s)) == Path("~\\bar")


bear = "🐻"  # Unicode character: "Bear Face" (U+1F43B)


@pytest.mark.parametrize(
    ("args", "expected"),
    [
        pytest.param(
            (alphabet, 255, True),
            alphabet,
            id="text - no truncation",
        ),
        pytest.param(
            (alphabet * 10, 255, True),
            (alphabet * 10)[:255],
            id="text - truncate",
        ),
        pytest.param(
            (alphabet * 10, 50, True),
            (alphabet * 10)[:50],
            id="text - truncate at 50",
        ),
        pytest.param(
            (f"{alphabet}.ext", 255, True),
            f"{alphabet}.ext",
            id="text+ext1 - no truncation",
        ),
        pytest.param(
            (f"{alphabet * 10}.ext", 255, True),
            f"{(alphabet * 10)[:251]}.ext",
            id="text+ext1 - truncate",
        ),
        pytest.param(
            (f"{alphabet * 10}.ext", 50, True),
            f"{(alphabet * 10)[:46]}.ext",
            id="text+ext1 - truncate at 50",
        ),
        pytest.param(
            (f"{alphabet * 10}.ext", 255, False),
            (alphabet * 10)[:255],
            id="text+ext1+nokeep - truncate",
        ),
        pytest.param(
            (f"{alphabet * 10}.ext", 50, False),
            (alphabet * 10)[:50],
            id="text+ext1+nokeep - truncate at 50",
        ),
        pytest.param(
            (f"{alphabet * 10}.notafilenameextension", 255, True),
            (alphabet * 10)[:255],
            id="text+ext2 - truncate",
        ),
        pytest.param(
            (f"{alphabet * 10}.notafilenameextension", 50, True),
            (alphabet * 10)[:50],
            id="text+ext2 - truncate at 50",
        ),
        pytest.param(
            (bear * 63, 255, True),
            bear * 63,
            id="bear - no truncation",
        ),
        pytest.param(
            (bear * 64, 255, True),
            bear * 63,
            id="bear - truncate",
        ),
        pytest.param(
            (bear * 64, 50, True),
            bear * 12,
            id="bear - truncate at 50",
        ),
        pytest.param(
            (f"{bear}.ext", 255, True),
            f"{bear}.ext",
            id="bear+ext1 - no truncation",
        ),
        pytest.param(
            (f"{bear * 64}.ext", 255, True),
            f"{bear * 62}.ext",
            id="bear+ext1 - truncate",
        ),
        pytest.param(
            (f"{bear * 64}.ext", 50, True),
            f"{bear * 11}.ext",
            id="bear+ext1 - truncate at 50",
        ),
        pytest.param(
            (f"{bear * 64}.ext", 255, False),
            bear * 63,
            id="bear+ext1+nokeep - truncate",
        ),
        pytest.param(
            (f"{bear * 64}.ext", 50, False),
            bear * 12,
            id="bear+ext1+nokeep - truncate at 50",
        ),
        pytest.param(
            (f"{bear * 64}.notafilenameextension", 255, True),
            bear * 63,
            id="bear+ext2 - truncate",
        ),
        pytest.param(
            (f"{bear * 64}.notafilenameextension", 50, True),
            bear * 12,
            id="bear+ext2 - truncate at 50",
        ),
    ],
)
def test_truncate_path(args: tuple[str, int, bool], expected: str):
    path, length, keep_extension = args
    result = truncate_path(path, length, keep_extension)
    assert len(result.encode("utf-8")) <= length
    assert result == expected



import re
from io import BytesIO

from streamlink import NoStreamsError
from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream
from streamlink.stream.stream import Stream


class TestStream(Stream):
    __shortname__ = "test"

    def open(self):
        return BytesIO(b"x" * 8192 * 2)


@pluginmatcher(
    re.compile(r"https?://test\.se"),
)
@pluginargument(
    "bool",
    action="store_true",
)
@pluginargument(
    "password",
    sensitive=True,
    metavar="PASSWORD",
)
class TestPlugin(Plugin):
    id = "test-id-1234-5678"
    author = "Tѥst Āuƭhǿr"
    category = None
    title = "Test Title"

    def _get_streams(self):
        if "empty" in self.url:
            return

        if "UnsortableStreamNames" in self.url:

            def gen():
                for _ in range(3):
                    yield "vod", HTTPStream(self.session, "http://test.se/stream")

            return gen()

        if "NoStreamsError" in self.url:
            raise NoStreamsError

        if "fromoptions" in self.url:
            return {"fromoptions": HTTPStream(self.session, self.options.get("streamurl"))}

        streams = {}
        streams["test"] = TestStream(self.session)
        streams["hls"] = HLSStream(self.session, "http://test.se/playlist.m3u8")
        streams["http"] = HTTPStream(self.session, "http://test.se/stream")

        streams["240p"] = HTTPStream(self.session, "http://test.se/stream")
        streams["360p"] = HTTPStream(self.session, "http://test.se/stream")
        streams["1080p"] = HTTPStream(self.session, "http://test.se/stream")

        streams["350k"] = HTTPStream(self.session, "http://test.se/stream")
        streams["800k"] = HTTPStream(self.session, "http://test.se/stream")
        streams["1500k"] = HTTPStream(self.session, "http://test.se/stream")
        streams["3000k"] = HTTPStream(self.session, "http://test.se/stream")

        streams["480p"] = [
            HTTPStream(self.session, "http://test.se/stream"),
            HLSStream(self.session, "http://test.se/playlist.m3u8"),
        ]

        return streams


__plugin__ = TestPlugin



from streamlink.plugins.chzzk import Chzzk
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlChzzk(PluginCanHandleUrl):
    __plugin__ = Chzzk

    should_match_groups = [
        (("live", "https://chzzk.naver.com/live/CHANNEL_ID"), {"channel_id": "CHANNEL_ID"}),
        (("video", "https://chzzk.naver.com/video/VIDEO_ID"), {"video_id": "VIDEO_ID"}),
        (("clip", "https://chzzk.naver.com/clips/CLIP_ID"), {"clip_id": "CLIP_ID"}),
    ]
